#! /usr/bin/env python
# -*- coding: utf-8 -*-


import numpy as np
import astropy.units as u
from p_winds import tools


_TEST_URL = 'https://raw.githubusercontent.com/ladsantos/p-winds/main/data/solar_spectrum_scaled_lambda.dat'


# Test the nearest_index function
def test_nearest_index():
    array = np.linspace(1, 10, 10)
    ind = tools.nearest_index(array, np.pi)
    assert ind == 2


# Test the make_spectrum_from_file function
def test_make_spectrum_from_file(precision_threshold=1E-6):
    units = {'wavelength': u.angstrom, 'flux': u.erg / u.s / u.cm ** 2 /
                                               u.angstrom}
    spectrum = tools.make_spectrum_from_file(_TEST_URL, units)
    test_value = (spectrum['flux_lambda'][0] * spectrum['flux_unit']).to(
        u.W / u.m ** 2 / u.angstrom).value
    assert abs((test_value - 2.08405464e-11) / test_value) < precision_threshold


# Test the standard spectrum function
_SPEC_TEST_URL = 'https://gist.githubusercontent.com/ladsantos/2cd9125a4efd8fff0f9b5967beafebf6/raw/4aee1bd9acda18c3aa95f3ce2ceb62d333dded6b/'
def test_standard_spectrum(precision_threshold=1E-6):
    spectrum = tools.standard_spectrum('solar', semi_major_axis=2.0,
                                       reference_spectra_dir=_SPEC_TEST_URL)
    test_value = (spectrum['flux_lambda'][0] * spectrum['flux_unit']).to(
        u.W / u.m ** 2 / u.angstrom).value
    assert abs((test_value - 5.33758E-10) / test_value) < precision_threshold


# Test the generate_muscles_spectrum function
def test_generate_muscles_spectrum(precision_threshold=1E-6):
    spectrum = tools.generate_muscles_spectrum('gj436',
                                               semi_major_axis=2.0,
                                               reference_spectra_dir='https://archive.stsci.edu/missions/hlsp/muscles/gj436/')
    test_value = (spectrum['flux_lambda'][0] * spectrum['flux_unit']).to(
        u.W / u.m ** 2 / u.angstrom).value
    assert abs((test_value - 0.0170225) / test_value) < precision_threshold



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from p_winds import lines, transit


data_test_url = 'https://raw.githubusercontent.com/ladsantos/p-winds/main/data/he_3_profile.dat'


r_pl = 1.39 * 71492000  # Planet radius in m
r = np.loadtxt(data_test_url, usecols=(0,))  # Altitudes in m
v = np.loadtxt(data_test_url, usecols=(1,))  # Velocities in m / s
n_he_3 = np.loadtxt(data_test_url, usecols=(2, ))  # He fraction
planet_to_star_ratio = 0.12086
w0, w1, w2, f0, f1, f2, a_ij = lines.he_3_properties()
m_He = 4 * 1.67262192369e-27  # Helium atomic mass in kg
wl = np.linspace(1.0827, 1.0832, 150) * 1E-6  # Wavelengths in m
v_wind = -2E3  # Line-of-sight wind velocity in m / s
w_array = np.array([w0, w1, w2])
f_array = np.array([f0, f1, f2])
a_array = np.array([a_ij, a_ij, a_ij])
t_0 = 9E3  # Upper-atmospheric temperature in K


# Test both the draw_transit and radiative_transfer functions
def test_draw_transit_and_radiative_transfer(precision_threshold=5E-3):
    flux_map, t_depth, r_from_planet = transit.draw_transit(
        planet_to_star_ratio=planet_to_star_ratio,
        planet_physical_radius=r_pl,
        impact_parameter=0.0,
        phase=0.0,
        supersampling=10,
        grid_size=101
    )
    test_value_0 = flux_map[30, 30]
    assert abs((test_value_0 - 1.249219E-4) / test_value_0) < \
        precision_threshold
    assert abs((t_depth - 0.015012130070238605) / t_depth) < \
           precision_threshold

    spectrum = transit.radiative_transfer_2d(flux_map, r_from_planet,
                                             r, n_he_3, v, w_array, f_array,
                                             a_array, wl, t_0, m_He,
                                             bulk_los_velocity=v_wind,
                                             wind_broadening_method='average')
    test_value_2 = spectrum[60]
    assert abs((test_value_2 - 0.97946) / test_value_2) < precision_threshold



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module is used to compute a grid containing drawing of disks of stars,
planets and atmospheres.
"""

import numpy as np
from scipy.interpolate import interp1d
from scipy.special import voigt_profile
from scipy.integrate import trapz
from flatstar import draw, utils


__all__ = ["draw_transit", "radiative_transfer_2d", "profile_los",
           "optical_depth_2d"]


# Draw a grid
def draw_transit(planet_to_star_ratio, planet_physical_radius,
                 impact_parameter=0.0, phase=0.0, grid_size=100,
                 supersampling=None, resample_method=None,
                 limb_darkening_law=None, ld_coefficient=None):
    """
    Calculate a normalized transit map. Additionally, calculate two-dimensional
    arrays of density and line-of-sight radial velocities around the planet.

    Parameters
    ----------
    planet_to_star_ratio : ``float``
        Ratio between the radii of the planet and the star.

    planet_physical_radius : ``float``
        Physical radius of the planet in whatever unit you want to work with.

    impact_parameter : ``float``, optional
        Transit impact parameter of the planet (not the atmosphere). Default is
        0.0.

    phase : ``float``, optional
        Phase of the transit. -0.5, 0.0, and +0.5 correspond to the center of
        planet being located at, respectively, the left limb of the star, the
        center, and the right limb. Default is 0.0.

    grid_size : ``int``, optional
        Size of the transit grid. Default is 100.

    supersampling : ``float`` or ``None``, optional
        In order to avoid pixels with hard edges, it is useful to first compute
        the transit grid at a high resolution and then downscale it to a
        manageable grid size. Supersampling is the factor by which to increase
        the grid size at first and then downscale to the requested grid size. If
        ``None``, no supersampling is applied. Default is ``None``.

    resample_method : ``str`` or ``None``, optional
        Method by which to resample the image if supersampling is used. If
        ``None``, then fallback to a "box" method. Available methods are
        ``"nearest"``, ``"box"``, ``"bilinear"``, ``"hamming"`` and
        ``"lanczos"`` (the last two are not recommended for research-grade
        results. Default is ``None``.

    limb_darkening_law : ``None`` or ``str``, optional
        String with the name of the limb-darkening law. The options are the same
        currently implemented in the code ``flatstar``: ``'linear'``,
        ``'quadratic'``, ``'square-root'``, ``'log'``, ``'exp'``, ``'s3'``,
        ``'c4'``, or ``None`` (no limb-darkening). Default is ``None``.

    ld_coefficient : ``float`` or ``array-like``
        In case of a linear limb-darkening law, the value of the coefficient
        should be a float. In all other options it should be array-like. Default
        is ``None``.

    Returns
    -------
    normalized_intensity_map : ``numpy.ndarray``
        2-D map of intensity normalized in such a way that the sum of the array
        will be 1.0 if the planet is not transiting.

    transit_depth : ``float``
        Absorption caused by the opaque disk of the planet in the specified
        transit configuration.

    r_from_planet : ``numpy.ndarray``
        2-D map of radial distances of each pixel from the center of the planet
        in the same unit as ``planet_physical_radius``.
    """
    if supersampling is not None:
        effective_grid_size = int(round(grid_size * supersampling))
        rescaling = 1 / supersampling
    else:
        effective_grid_size = grid_size
        rescaling = None

    star_grid = draw.star(effective_grid_size,
                          limb_darkening_law=limb_darkening_law,
                          ld_coefficient=ld_coefficient)
    transit_grid = draw.planet_transit(star_grid, planet_to_star_ratio,
                                       impact_parameter, phase,
                                       rescaling_factor=rescaling,
                                       resample_method=resample_method)

    # We will calculate the densities and wind radial velocities of the
    # upper atmosphere. We need to know the matrix r_p containing distances from
    # planet center when we draw the extended atmosphere
    pl_ref = transit_grid.planet_px_coordinates
    planet_centric_r = utils.cylindrical_r(transit_grid.intensity, pl_ref)
    # We also need to know the physical size of the pixel in the grid
    planet_radius = transit_grid.planet_radius_px
    px_size = planet_physical_radius / planet_radius
    r_from_planet = planet_centric_r * px_size

    # Finally
    normalized_intensity_map = transit_grid.intensity
    transit_depth = transit_grid.transit_depth

    return normalized_intensity_map, transit_depth, r_from_planet


# Calculate the radiative transfer
def radiative_transfer_2d(intensity_0, r_from_planet, radius_profile,
                          density_profile, velocity_profile, central_wavelength,
                          oscillator_strength, einstein_coefficient,
                          wavelength_grid, gas_temperature, particle_mass,
                          bulk_los_velocity=0.0, planet_radial_velocity=0.0,
                          wind_broadening_method='average', z_grid_size=200,
                          turbulence_broadening=False):
    """
    Calculate the absorbed intensity profile in a wavelength grid.

    Parameters
    ----------
    intensity_0 : ``float`` or ``numpy.ndarray``
        Original flux intensity originating from a background illuminating
        source. If ``numpy.ndarray``, must have the same shape as
        ``r_from_planet``.

    r_from_planet : ``numpy.ndarray``
        2-D map of radial distances of each pixel from the center of the planet.
        The unit has to be consistent with the other input parameters. E.g., if
        you use unit of meters, all other input parameters with length in the
        unit also have to be in meters.

    radius_profile : ``numpy.ndarray``
        1-D profile of radii in which the densities are sampled. Unit has to be
        the same as ``r_from_planet``.

    density_profile : ``numpy.ndarray``
        1-D profile of volumetric number densities in function of radius. Unit
        has to be 1 / length ** 3, where length is the same unit as
        ``r_from_planet``.

    velocity_profile : ``numpy.ndarray``
        1-D profile of velocities in function of radius. Unit has to be
        m / s.

    central_wavelength : ``float`` or ``array_like``
        Central wavelength of the transition in unit of m. If more than one line
        is to be calculated, then input this parameter as an array-like object.

    oscillator_strength : ``float``
        Oscillator strength of the transition. The format or shape of this input
        parameter needs to be consistent with that of ``central_wavelength``.

    einstein_coefficient : ``float``
        Einstein coefficient of the transition in 1 / s. The format or shape of
        this input parameter needs to be consistent with that of
        ``central_wavelength``.

    wavelength_grid : ``numpy.ndarray``
        Wavelengths to calculate the profile in unit of m.

    gas_temperature : ``float`` or ``numpy.ndarray``
        1-D profile of temperatures in function of radius. Unit has to be K. If
        defined as ``float``, then the model assumes isothermal gas.

    particle_mass : ``float``
        Mass of the particle corresponding to the transition in unit of kg.

    bulk_los_velocity : ``float``, optional
        Bulk velocity of the gas cell in the line of sight in unit of m / s.
        Default is 0.0.

    planet_radial_velocity : ``float``, optional
        Radial velocity (i.e., the component in the line of sight) of the planet
        in relation to the rest frame of the host star and in unit of m / s.
        Default is 0.0.

    wind_broadening_method : ``str``, optional
        Method of calculation for the wind broadening. There are two options:
        1) ``'formal'``: the formal definition of radiative transfer taking into
        account the full dimensionality of the wind (slower);
        2) ``'average'``: assumes the Parker wind broadening contributes to the
        Gaussian term of the Voigt profile with an additive factor proportional
        to the square of the density-averaged, line-of-sight velocity (faster).

    z_grid_size : ``int``, optional
        Grid size for the line of sight direction. This is used only if
        ``wind_broadening_method`` is set to ``'formal'``. Default is 200.

    turbulence_broadening : ``bool``, optional
        If ``True``, adds a turbulence broadening, defined as in Lampón et al.
        2020, to the Gaussian term of the Voigt profile. Default is ``False``.

    Returns
    -------
    intensity : ``numpy.ndarray``
        Absorbed intensity profile in function of ``wavelength_grid``.
    """
    # First, calculate the optical depth in function of radial distance from the
    # planet and the wavelength
    optical_depth_profile = optical_depth_2d(
        radius_profile, density_profile, velocity_profile, central_wavelength,
        oscillator_strength, einstein_coefficient, wavelength_grid,
        gas_temperature, particle_mass, z_grid_size, bulk_los_velocity,
        planet_radial_velocity, wind_broadening_method,
        turbulence_broadening
    )

    # Now we interpolate the optical depths to each radius from the planet
    f = interp1d(radius_profile, optical_depth_profile, axis=0,
                 bounds_error=False, fill_value=0.0)
    optical_depth_array = f(r_from_planet)

    # A bit of array-manipulation magic here to allow us to properly broadcast
    intensity_0 = np.reshape(intensity_0, np.shape(intensity_0) + (1,))

    # Finally calculate the radiative transfer
    intensity = np.sum(intensity_0 * np.exp(-optical_depth_array), axis=(0, 1))

    return intensity


# Density and velocity profiles in the line of sight
def profile_los(radius_profile, density_profile, velocity_profile,
                z_grid_size, temperature_profile=None):
    """
    Calculate the profiles of radius and line-of-sight velocities in function of
    sky-projected radial distance from the planet (axis 0) and the line of sight
    direction (axis 1).

    Parameters
    ----------
    radius_profile : ``numpy.ndarray``
        1-D profile of radii in which the densities are sampled, in whatever
        unit you want to work with.

    density_profile : ``numpy.ndarray``
        1-D profile of volumetric number densities in function of radius, in
        whatever unit you want to work with.

    velocity_profile : ``numpy.ndarray``
        1-D profile of velocities in function of radius, in whatever unit you
        want to work with.

    z_grid_size : ``int``, optional
        Grid size for the line of sight direction.

    temperature_profile : ``numpy.ndarray``, optional
        1-D profile of temperatures in function of radius, in whatever unit you
        want to work with. If ``None``, then the code assumes isothermal
        temperature. Default is ``None``.

    Returns
    -------
    los_density_r_z : ``numpy.ndarray``
        2-D map of densities in the x- and z-axis distances from the center of
        the planet.

    los_velocity_r_z : ``numpy.ndarray``
        2-D map of line-of-sight velocities in the x- and z-axis distances from
        the center of the planet.
    """
    # Create a two-dimensional array that measures the distance from the planet
    # The second dimension is the line-of-sight direction.
    r_top_atm = radius_profile[-1]
    los_z = np.linspace(-r_top_atm, r_top_atm, z_grid_size)
    # Maybe in the future implement z in log-space instead of linear space

    coordinates = np.array(np.meshgrid(radius_profile, los_z))
    distances = np.sum(coordinates ** 2, axis=0) ** 0.5

    # The line-of-sight wind velocity is given by Eq. 6 in Seidel et al.
    # (2020)
    # (https://ui.adsabs.harvard.edu/abs/2020A%26A...633A..86S/abstract)
    f_v = interp1d(radius_profile, velocity_profile, bounds_error=False,
                   fill_value=0.0)
    v_vertical = f_v(distances)
    # We need to add one dimension to `z` to allow for proper array broadcasting
    z_expanded = np.expand_dims(los_z, axis=1)
    los_velocity_r_z = v_vertical * z_expanded / \
        (z_expanded ** 2 + radius_profile ** 2) ** 0.5

    # Calculate the line-of-sight density
    d_v = interp1d(radius_profile, density_profile, bounds_error=False,
                   fill_value=0.0)
    los_density_r_z = d_v(distances)

    if temperature_profile is not None:
        # Calculate the line-of-sight temperature
        t_v = interp1d(radius_profile, temperature_profile, bounds_error=False,
                       fill_value=0.0)
        los_temperature_r_z = t_v(distances)

        return los_density_r_z, los_velocity_r_z, los_temperature_r_z, los_z

    else:
        return los_density_r_z, los_velocity_r_z, los_z


# Optical depth in function of cylindrical radius from the planet and
# wavelength. Hold on to your hat because this code is very complex.
def optical_depth_2d(radius_profile, density_profile, velocity_profile,
                     central_wavelength, oscillator_strength,
                     einstein_coefficient, wavelength_grid, gas_temperature,
                     particle_mass, z_grid_size, bulk_los_velocity=0.0,
                     planet_radial_velocity=0.0,
                     wind_broadening_method='average',
                     turbulence_broadening=False):
    """
    Calculate the optical depth in function of cylindrical radius from the
    planet and the wavelength.

    Parameters
    ----------
    radius_profile : ``numpy.ndarray``
        1-D profile of radii in which the densities are sampled. Unit has to be
        consistent with the other input parameters involving lenghts and
        densities.

    density_profile : ``numpy.ndarray``
        1-D profile of volumetric number densities in function of radius. Unit
        has to be 1 / length ** 3, where length is the same unit as
        ``radius_profile``.

    velocity_profile : ``numpy.ndarray``
        1-D profile of velocities in function of radius. Unit has to be
        m / s.

    central_wavelength : ``float`` or ``array_like``
        Central wavelength of the transition in unit of m. If more than one line
        is to be calculated, then input this parameter as an array-like object.

    oscillator_strength : ``float``
        Oscillator strength of the transition. The format or shape of this input
        parameter needs to be consistent with that of ``central_wavelength``.

    einstein_coefficient : ``float``
        Einstein coefficient of the transition in 1 / s. The format or shape of
        this input parameter needs to be consistent with that of
        ``central_wavelength``.

    wavelength_grid : ``numpy.ndarray``
        Wavelengths to calculate the profile in unit of m.

    gas_temperature : ``float`` or ``numpy.ndarray``
        1-D profile of temperatures in function of radius. Unit has to be K. If
        defined as ``float``, then the model assumes isothermal gas.

    particle_mass : ``float``
        Mass of the particle corresponding to the transition in unit of kg.

    z_grid_size : ``int``
        Grid size for the line of sight direction.

    bulk_los_velocity : ``float``, optional
        Bulk velocity of the gas cell in the line of sight in unit of m / s.
        Default is 0.0.

    planet_radial_velocity : ``float``, optional
        Radial velocity (i.e., the component in the line of sight) of the planet
        in relation to the rest frame of the host star and in unit of m / s.
        Default is 0.0.

    wind_broadening_method : ``str``, optional
        Method of calculation for the wind broadening. There are two options:
        1) ``'formal'``: the formal definition of radiative transfer taking into
        account the full dimensionality of the wind (slower);
        2) ``'average'``: assumes the Parker wind broadening contributes to the
        Gaussian term of the Voigt profile with an additive factor proportional
        to the square of the density-averaged, line-of-sight velocity (faster).

    turbulence_broadening : ``bool``, optional
        If ``True``, adds a turbulence broadening, defined as in Lampón et al.
        2020, to the Gaussian term of the Voigt profile. Only used if
        ``wind_broadening_method`` is set to ``'average'``. Default is ``False``.

    Returns
    -------
    optical_depth_array : ``numpy.ndarray``
        Optical depth in function of radial distance from the center of the
        planet (axis 0) and in function of wavelength (axis 1).
    """
    # Calculate density, velocity profiles and, if necessary, temperature
    # profile in the line of sight
    if isinstance(gas_temperature, float) or isinstance(gas_temperature, int):
        density_los, velocity_los, z_los = \
            profile_los(radius_profile, density_profile, velocity_profile,
                        z_grid_size)
        spatial_shape = np.shape(density_los)
        temp = gas_temperature
    elif isinstance(gas_temperature, np.ndarray):
        density_los, velocity_los, temperature_los, z_los = \
            profile_los(radius_profile, density_profile, velocity_profile,
                        z_grid_size=z_grid_size,
                        temperature_profile=gas_temperature)
        temp = temperature_los
    else:
        raise ValueError('`gas_temperature` has to be either `float` or '
                         '`np.ndarray`.')

    # Spectral line properties
    w0 = central_wavelength  # Reference wavelength in m
    f = oscillator_strength  # Unit-less
    a_ij = einstein_coefficient  # In unit of 1 / s

    # Transform `w0`, `f` and `a_ij` into arrays if they are not already arrays
    if isinstance(w0, float):
        w0 = np.array([w0, ])
        f = np.array([f, ])
        a_ij = np.array([a_ij, ])
    elif isinstance(w0, list):
        w0 = np.array(w0)
        f = np.array(f)
        a_ij = np.array(a_ij)
    elif isinstance(w0, np.ndarray):
        pass
    else:
        raise ValueError('The spectral line properties must be float, list or'
                         ' numpy.ndarray, and their format must be consistent'
                         ' with each other.')

    # Some necessary book-keeping here
    n_lines = len(w0)
    wl_grid = wavelength_grid
    c_speed = 2.99792458e+08  # Speed of light in m / s
    k_b = 1.380649e-23  # Boltzmann's constant in J / K
    nu0 = c_speed / w0  # Reference frequency in Hz
    nu_grid_rest = c_speed / wl_grid
    v_bulk = bulk_los_velocity
    rv_planet = planet_radial_velocity
    mass = particle_mass

    # Change the shape of `nu_grid_rest` to allow for proper array broadcasting
    nu_grid_rest = np.reshape(nu_grid_rest, (len(nu_grid_rest), 1))

    # For each line we calculate the grid of frequency shifts from the central
    # frequency
    delta_nu = (nu_grid_rest - nu0).T

    # Cross-section
    k = 2.654008854574474e-06  # Physical constant in units of m ** 2 * Hz
    sigma = k * f

    # This is convoluted, but we create a nested function to calculate the
    # optical depth divided by the cross-section (i.e. normalized optical depth)
    def _normalized_optical_depth(delta_nu_grid, nu0_k, a_ij_k,
                                  _method=wind_broadening_method):
        # Calculate the Lorentzian width of the Voigt profile
        gamma = a_ij_k / 4 / np.pi

        # Formal calculation of optical depth (slower)
        if _method == 'formal':
            # Calculate Doppler width (standard deviation) of the Voigt profile
            alpha_nu = \
                nu0_k / c_speed * (k_b * temp / mass) ** 0.5

            # Calculate the frequency shifts due to wind and bulk motion
            delta_nu_wind = (velocity_los + v_bulk +
                             rv_planet) / c_speed * nu0_k
            delta_nu_add = np.reshape(delta_nu_wind, spatial_shape + (1,))

        # Faster calculation of optical depth: We assume that the Parker wind
        # broadening has a Gaussian shape. To this end, we take a wind velocity
        # and add it quadratically to the Gaussian broadening term of the Voigt
        # profile.
        elif _method == 'average':
            # Calculate the wind broadening velocity as the density-averaged
            # line-of-sight velocity of the Parker wind
            weights = density_los
            wind_broadening_velocity = \
                (np.sum(velocity_los ** 2 * weights) /
                 np.sum(weights)) ** 0.5

            if isinstance(temp, np.ndarray):
                average_temp = np.sum(temp * weights) / np.sum(weights)
            elif isinstance(temp, float) or isinstance(temp, int):
                average_temp = temp

            if turbulence_broadening is True:
                # Similar to Lampon et al. (2020), calculate the broadening
                # as the turbulent velocity = sqrt(5/3 * kT / 2m)
                turbulence_velocity = np.sqrt(5 / 6 * k_b * average_temp / mass)
            else:
                turbulence_velocity = 0.0

            # Calculate Doppler width of the Voigt profile
            alpha_nu = nu0_k / c_speed * (k_b * average_temp / mass +
                                          wind_broadening_velocity ** 2 +
                                          turbulence_velocity ** 2) ** 0.5

            # Frequency shift due to bulk line-of-sight velocity (not to be
            # confused with the Parker wind velocity).
            delta_nu_add = (v_bulk + rv_planet) / c_speed * nu0_k

        else:
            raise ValueError('The chosen ``wind_broadening_method`` is not '
                             'implemented.')

        # Finally calculate the Voigt profiles
        profiles = voigt_profile(delta_nu_grid + delta_nu_add, alpha_nu,
                                 gamma)

        # Calculate the optical depths divided by the cross section
        density_expanded = np.expand_dims(density_los, axis=-1)
        opt_depth_over_cross_section_r_nu = \
            trapz(profiles * density_expanded, z_los, axis=0)

        return opt_depth_over_cross_section_r_nu

    # Calculate the optical depth for each line
    optical_depth_list = []
    for i in range(n_lines):
        norm_opt_depth = _normalized_optical_depth(
            delta_nu_grid=delta_nu[i],
            nu0_k=nu0[i],
            a_ij_k=a_ij[i]
        )
        optical_depth_i = norm_opt_depth * sigma[i]
        optical_depth_list.append(optical_depth_i)
    optical_depth_array = np.array(optical_depth_list)
    optical_depth_array = np.sum(optical_depth_array, axis=0)

    return optical_depth_array



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module contains useful tools to facilitate numerical calculations.
"""

from __future__ import (division, print_function, absolute_import,
                        unicode_literals)
import numpy as np
import astropy.units as u
import os
from warnings import warn
from astropy.io import fits


__all__ = ["nearest_index", "standard_spectrum", "generate_muscles_spectrum",
           "make_spectrum_from_file"]

# Find $PWINDS_REFSPEC_DIR environment variable
try:
    _PWINDS_REFSPEC_DIR = os.environ["PWINDS_REFSPEC_DIR"]
except KeyError:
    _PWINDS_REFSPEC_DIR = None
    warn("Environment variable PWINDS_REFSPEC_DIR is not set.")


def nearest_index(array, target_value):
    """
    Finds the index of a value in ``array`` that is closest to ``target_value``.

    Parameters
    ----------
    array : ``numpy.array``
        Target array.
    target_value : ``float``
        Target value.

    Returns
    -------
    index : ``int``
        Index of the value in ``array`` that is closest to ``target_value``.
    """
    index = array.searchsorted(target_value)
    index = np.clip(index, 1, len(array) - 1)
    left = array[index - 1]
    right = array[index]
    index -= target_value - left < right - target_value
    return index


def standard_spectrum(stellar_type, semi_major_axis,
                      reference_spectra_dir=_PWINDS_REFSPEC_DIR,
                      stellar_radius=None, truncate_wavelength_grid=False,
                      cutoff_thresh=13.6):
    """
    Construct a dictionary containing an input spectrum for a given spectral
    type. The code scales this to the spectrum received at your planet provided
    a value for the scaled ``semi_major_axis``.

    Spectrum of iota Horologii was kindly provided by Jorge Sanz-Forcada (priv.
    comm.). Spectrum of HD 108147 and HR 8799 were obtained from the
    X-exoplanets database and combined with PHOENIX atmosphere models for the
    NUV. Solar spectrum comes from the Whole Heliosphere Interval (WHI)
    Reference Spectra obtained from the LASP Interactive Solar Irradiance
    Datacenter. All other spectra are from the MUSCLES survey.

    Parameters
    ----------
    stellar_type : ``str``
        Define the stellar type. The available options are:

        - ``'mid-A'`` (based on HR 8799)
        - ``'early-F'`` (based on WASP-17)
        - ``'late-F'`` (based on HD 108147)
        - ``'early-G'`` (based on HD 149026)
        - ``'solar'`` (based on the Sun)
        - ``'young-sun'`` (based on iota Horologii)
        - ``'late-G'`` (based on TOI-193)
        - ``'active-K'`` (based on epsilon Eridanii)
        - ``'early-K'`` (based on HD 97658)
        - ``'late-k'`` (based on WASP-43)
        - ``'active-M'`` (based on Proxima Centauri)
        - ``'early-M'`` (based on GJ 436)
        - ``'late-M'`` (based on TRAPPIST-1)

    semi_major_axis : ``float``
        Semi-major axis of the planet in units of stellar radii. The code first
        converts the MUSCLES spectrum to what it would be at R_star;
        ``semi_major_axis`` is needed to get the spectrum at the planet.

    reference_spectra_dir : ``str``, optional
        Path to the directory with the MUSCLES data. Default value is defined by
        the environment variable ``$PWINDS_REFSPEC_DIR``.

    stellar_radius : ``float``, optional
        Stellar radius in unit of solar radii. Setting a value for this
        parameter allows the spectrum to be scaled to an arbitrary stellar
        radius instead of the radius of the MUSCLES star. If ``None``, then the
        scaling is performed using the radius of the MUSCLES star. Default is
        ``None``.

    truncate_wavelength_grid : ``bool``, optional
        If ``True``, will only return the spectrum with energy > 13.6 eV. This
        may be useful for computational expediency. If False, returns the whole
        spectrum. Default is ``False``.

    cutoff_thresh : ``float``, optional
        If ``truncate_wavelength_grid`` is set to ``True``, then the truncation
        happens for energies whose value in eV is above this threshold, also in
        eV. Default is ``13.6``.

    Returns
    -------
    spectrum : ``dict``
        Spectrum dictionary with entries for the wavelength and flux, and their
        units.
    """
    muscles_match = {'early-A': None, 'late-A': None, 'early-F': 'wasp-17',
                     'late-F': None, 'early-G': 'hd-149026',
                     'late-G': 'toi-193', 'solar': None, 'young-Sun': None,
                     'active-K': 'v-eps-eri', 'early-K': 'hd97658',
                     'late-K': 'wasp-43', 'active-M': 'gj551',
                     'early-M': 'gj436', 'late-M': 'trappist-1'}

    try:
        spectrum = generate_muscles_spectrum(muscles_match[stellar_type],
                                             semi_major_axis,
                                             reference_spectra_dir,
                                             stellar_radius,
                                             truncate_wavelength_grid,
                                             cutoff_thresh)
    except KeyError:
        prefix = reference_spectra_dir

        # Check if prefix has a trailing forward slash
        if prefix[-1] == '/':
            pass
        # If not, add it
        else:
            prefix += '/'

        if stellar_type == 'solar':
            spectrum_array = np.loadtxt(
                prefix + 'ref_solar_irradiance_whi-2008_ver2.dat', skiprows=142,
                usecols=(0, 2))
            i1 = nearest_index(spectrum_array[:, 0], 300)
            wavelength = (spectrum_array[:i1, 0] * u.nm).to(u.angstrom).value
            flux = (spectrum_array[:i1, 1] * u.W / u.m ** 2 / u.nm).to(
                u.erg / u.s / u.cm ** 2 / u.angstrom).value
            r_star_origin = 1.00 * u.solRad
            dist = 1 * u.au
        elif stellar_type == 'young-Sun':
            spectrum_array = np.loadtxt(prefix + 'spec_hr810_1au.dat')
            wavelength = spectrum_array[:, 0]
            flux = spectrum_array[:, 1]
            r_star_origin = 1.00 * u.solRad  # Assumption
            dist = 1 * u.au
        elif stellar_type == 'mid-A':
            spectrum_array = np.loadtxt(prefix + 'spec_hr8799_1au.dat')
            wavelength = spectrum_array[:, 0]
            flux = spectrum_array[:, 1]
            r_star_origin = 1.44 * u.solRad  # From Gaia DR2 for HR 8799
            dist = 1 * u.au
        elif stellar_type == 'late-F':
            spectrum_array = np.loadtxt(prefix + 'spec_hd108147_1au.dat')
            wavelength = spectrum_array[:, 0]
            flux = spectrum_array[:, 1]
            r_star_origin = 1.23 * u.solRad  # From Gaia DR2 for HD 108147
            dist = 1 * u.au
        else:
            raise ValueError('Specified stellar type not recognized')

        if stellar_radius is None:
            r_star = r_star_origin
        else:
            r_star = stellar_radius * u.solRad

        conv = float((dist / r_star) ** 2)  # conversion to
        # spectrum at R_star
        spectrum = {'wavelength': wavelength,
                    'flux_lambda': flux * conv * semi_major_axis ** (-2),
                    'wavelength_unit': u.AA,
                    'flux_unit': u.erg / u.s / u.cm / u.cm / u.AA}

    return spectrum


def generate_muscles_spectrum(host_star_name, semi_major_axis,
                              reference_spectra_dir=_PWINDS_REFSPEC_DIR,
                              stellar_radius=None,
                              truncate_wavelength_grid=False,
                              cutoff_thresh=13.6):
    """
    Construct a dictionary containing an input spectrum from a MUSCLES spectrum.
    MUSCLES reports spectra as observed at Earth, the code scales this to the
    spectrum received at your planet provided a value for the scaled
    ``semi-major-axis``.

    Parameters
    ----------
    host_star_name : ``str``
        Name of the MUSCLES stellar spectrum you want to use. Must be one of:
        ['gj176', 'gj436', 'gj551', 'gj581', 'gj667c', 'gj832', 'gj876',
        'gj1214', 'hd40307', 'hd85512', 'hd97658', 'v-eps-eri', 'gj1132',
        'hat-p-12', 'hat-p-26', 'hd-149026', 'l-98-59', 'l-678-39', 'l-980-5',
        'lhs-2686', 'lp-791-18', 'toi-193', 'trappist-1', 'wasp-17', 'wasp-43',
        'wasp-77a', 'wasp-127'].

    semi_major_axis : ``float``
        Semi-major axis of the planet in units of stellar radii. The code first
        converts the MUSCLES spectrum to what it would be at R_star;
        ``semi_major_axis`` is needed to get the spectrum at the planet.

    reference_spectra_dir : ``str``, optional
        Path to the directory with the reference spectra. Default value is
        defined by the environment variable ``$PWINDS_REFSPEC_DIR``.

    stellar_radius : ``float``, optional
        Stellar radius in unit of solar radii. Setting a value for this
        parameter allows the spectrum to be scaled to an arbitrary stellar
        radius instead of the radius of the MUSCLES star. If ``None``, then the
        scaling is performed using the radius of the MUSCLES star. Default is
        ``None``.

    truncate_wavelength_grid : ``bool``, optional
        If ``True``, will only return the spectrum with energy > 13.6 eV. This
        may be useful for computational expediency. If False, returns the whole
        spectrum. Default is ``False``.

    cutoff_thresh : ``float``, optional
        If ``truncate_wavelength_grid`` is set to ``True``, then the truncation
        happens for energies whose value in eV is above this threshold, also in
        eV. Default is ``13.6``.

    Returns
    -------
    spectrum : ``dict``
        Spectrum dictionary with entries for the wavelength and flux, and their
        units.
    """
    # Hard coding some values
    # The stellar radii and distances are taken from NASA Exoplanet Archive.

    thresh = cutoff_thresh * u.eV
    stars = [
        # Old ones
        'gj176', 'gj436', 'gj551', 'gj581', 'gj667c', 'gj832', 'gj876',
        'gj1214', 'hd40307', 'hd85512', 'hd97658', 'v-eps-eri',
        # New ones
        #'gj15a', 'gj163', 'gj649', 'gj674', 'gj676a', 'gj699', 'gj729', 'gj849',
        'gj1132', 'hat-p-12', 'hat-p-26', 'hd-149026', 'l-98-59', 'l-678-39',
        'l-980-5', 'lhs-2686', 'lp-791-18', 'toi-193', 'trappist-1', 'wasp-17',
        'wasp-43', 'wasp-77a', 'wasp-127'
        ]
    versions = np.array([
        # Old ones
        'v22', 'v22', 'v22', 'v22', 'v22', 'v22', 'v22',
        'v22', 'v22', 'v22', 'v22', 'v22',
        # New ones
        #'v23', 'v23', 'v23', 'v23', 'v23', 'v23', 'v23', 'v23',
        'v23', 'v24', 'v24', 'v24', 'v24', 'v24',
        'v23', 'v23', 'v24', 'v24', 'v23', 'v24',
        'v24', 'v24', 'v24'
    ])
    st_rads = np.array([
        # Old ones
        0.46, 0.449, 0.154, 0.3297020, 0.42, 0.45, 0.35, 0.22,
        0.71, 0.69, 0.74, 0.77,
        # New ones
        #
        0.21, 0.7, 0.87, 1.41, 0.3, 0.34,
        0.22,  # L 980-5 radius assumed to be the same as GJ 1214
        0.449,  # LHS 2686 radius assumed to be the same as GJ 436
        0.18, 0.95, 0.12, 1.49, 0.6, 0.910, 1.33
    ]) * u.solRad
    dists = np.array([
        # Old ones
        9.470450, 9.75321, 1.30119, 6.298100, 7.24396, 4.964350,
        4.67517, 14.6427, 12.9363, 11.2810, 21.5618,
        3.20260,
        # New ones
        #3.56244, 15.1353,
        12.613, 142.751, 141.837, 75.8643, 10.6194, 9.44181, 13.3731, 12.1893,
        26.4927, 80.4373, 12.4298888, 405.908, 86.7467, 105.6758, 159.507
    ]) * u.pc
    muscles_dists = {starname: dist for starname, dist in zip(stars, dists)}
    muscles_rstars = {starname: st_rad for starname, st_rad in zip(stars,
                                                                   st_rads)}
    muscles_versions = {starname: versions for starname, versions in zip(stars,
                                                                   versions)}

    # MUSCLES records spectra as observed at earth, so we need to convert it to
    # spectrum at R_star. The user has the option of setting an arbitary stellar
    # radius instead of the MUSCLES star radius to allow for more flexibility.
    # This can be especially useful for slightly evolved stars, whose radius
    # are larger than the MUSCLES stars.
    dist = muscles_dists[host_star_name]
    vnumber = muscles_versions[host_star_name]
    if stellar_radius is None:
        rstar = muscles_rstars[host_star_name]
    else:
        rstar = stellar_radius * u.solRad
    conv = float((dist / rstar) ** 2)  # conversion to spectrum at R_star

    # First check if reference_spectra_dir has a trailing forward slash
    if reference_spectra_dir[-1] == '/':
        pass
    # If not, add it
    else:
        reference_spectra_dir += '/'

    # Read the MUSCLES spectrum
    spec = fits.getdata(reference_spectra_dir +
                        f'hlsp_muscles_multi_multi_{host_star_name}_broadband_'
                        f'{vnumber}_adapt-const-res-sed.fits',
                        1)

    if truncate_wavelength_grid:
        mask = spec['WAVELENGTH'] * u.AA < thresh.to(u.AA,
                                                     equivalencies=u.spectral())
    else:
        mask = np.ones(spec.shape, dtype='bool')

    spectrum = {'wavelength': spec['WAVELENGTH'][mask],
                'flux_lambda': spec['FLUX'][mask] * conv *
                semi_major_axis ** (-2),
                'wavelength_unit': u.AA,
                'flux_unit': u.erg / u.s / u.cm / u.cm / u.AA}
    return spectrum


def make_spectrum_from_file(filename, units, path='', skiprows=0,
                            scale_flux=1.0, star_distance=None,
                            semi_major_axis=None):
    """
    Construct a dictionary containing an input spectrum from a text file. The
    input file must have two or more columns, in which the first is the
    wavelength or frequency bin center and the second is the flux per bin of
    wavelength or frequency. The code automatically figures out if the input
    spectra are binned in wavelength or frequency based on the units the user
    passes.

    Parameters
    ----------
    filename : ``str``
        Name of the file containing the spectrum data.

    units : ``dict``
        Units of the spectrum. This dictionary must have the entries
        ``'wavelength'`` and ``'flux'``, or ``'frequency'`` and ``'flux'``.
        The units must be set in ``astropy.units``.

    path : ``str``, optional
        Path to the spectrum data file.

    skiprows : ``int``, optional
        Number of rows to skip corresponding to the header of the input text
        file.

    scale_flux : ``float``, optional
        Scaling factor for flux. Default value is 1.0 (no scaling).

    star_distance : ``float`` or ``None``, optional
        Distance to star in unit of parsec. This is used to scale the flux as
        observed from Earth to the semi-major axis of the planet. If ``None``,
        no scaling is applied. If not ``None``, then a value``semi_major_axis``
        must be provided as well. Default is ``None``.

    semi_major_axis : ``float`` or ``None``, optional
        Semi-major axis of the planet in unit of au. This is used to scale the
        flux as observed from Earth to the semi-major axis of the planet. Notice
        that this parameter is different from the
        ``generate_muscles_spectrum()``  function, which uses the semi-major
        axis in unit of stellar radii. If ``None``, no scaling is applied. If
        not ``None``, then a value``star_distance`` must be provided as well.
        Default is ``None``.

    Returns
    -------
    spectrum : ``dict``
        Spectrum dictionary with entries for the wavelength and flux, and their
        units.

    """
    spectrum_table = np.loadtxt(path + filename, usecols=(0, 1),
                                skiprows=skiprows, dtype=float)

    try:
        x_axis = 'wavelength'
        x_axis_unit = units.pop(x_axis)
        y_axis = 'flux_lambda'
    except KeyError:
        x_axis = 'frequency'
        x_axis_unit = units.pop(x_axis)
        y_axis = 'flux_nu'
    y_axis_unit = units.pop('flux')

    conv_pc_to_au = 206264.8062471  # Conversion from pc to au
    if star_distance is not None and semi_major_axis is not None:
        scale_to_planet = \
            (star_distance * conv_pc_to_au / semi_major_axis) ** (-2)
    else:
        scale_to_planet = 1.0

    spectrum = {x_axis: spectrum_table[:, 0],
                y_axis: spectrum_table[:, 1] * scale_flux * scale_to_planet,
                '{}_unit'.format(x_axis): x_axis_unit,
                'flux_unit': y_axis_unit}
    return spectrum



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import astropy.units as u
from p_winds import hydrogen, helium, tools, parker, oxygen


# HD 209458 b
R_pl = (1.39 * u.jupiterRad).value
M_pl = (0.73 * u.jupiterMass).value
m_dot = (5E11 * u.g / u.s).value
T_0 = (9E3 * u.K).value
h_fraction = 0.90
he_fraction = 1 - h_fraction
he_h_fraction = he_fraction / h_fraction
average_f_ion = 0.0
average_mu = (1 + 4 * he_h_fraction) / (1 + he_h_fraction + average_f_ion)

r = np.logspace(0, np.log10(20), 100)  # Radius in unit of planetary radii
data_test_url = 'https://raw.githubusercontent.com/ladsantos/p-winds/main/data/solar_spectrum_scaled_lambda.dat'

units = {'wavelength': u.angstrom, 'flux': u.erg / u.s / u.cm ** 2 / u.angstrom}
spectrum = tools.make_spectrum_from_file(data_test_url, units)


# Let's test if the code is producing reasonable outputs.
def test_ion_fraction():
    # First calculate the hydrogen ion fraction
    f_r, mu_bar = hydrogen.ion_fraction(r, R_pl, T_0, h_fraction, m_dot, M_pl,
                                        average_mu, spectrum_at_planet=spectrum,
                                        relax_solution=True, return_mu=True)

    # Calculate the structure
    vs = parker.sound_speed(T_0, mu_bar)  # Speed of sound (km/s, assumed to be
    # constant)
    rs = parker.radius_sonic_point(M_pl, vs)  # Radius at the sonic point (jupiterRad)
    rhos = parker.density_sonic_point(m_dot, rs, vs)  # Density at the sonic point (g/cm^3)

    # Some useful arrays for the modeling
    r_array = r * R_pl / rs  # Radius in unit of radius at
    # sonic point
    v_array, rho_array = parker.structure(r_array)

    # Now calculate the population of helium
    f_he_plus = helium.ion_fraction(
        r, v_array, rho_array, f_r,
        R_pl, T_0, h_fraction, vs, rs, rhos, spectrum_at_planet=spectrum,
        initial_f_he_ion=0.0, relax_solution=False)

    f_o_ii = oxygen.ion_fraction(radius_profile=r,
                                 velocity=v_array,
                                 density=rho_array,
                                 hydrogen_ion_fraction=f_r,
                                 helium_ion_fraction=f_he_plus,
                                 planet_radius=R_pl,
                                 temperature=T_0,
                                 h_fraction=h_fraction,
                                 speed_sonic_point=vs,
                                 radius_sonic_point=rs,
                                 density_sonic_point=rhos,
                                 spectrum_at_planet=spectrum,
                                 initial_f_o_ion=0.0,
                                 relax_solution=True)

    # Assert if all values of the fractions are between 0 and 1
    n_neg = len(np.where(f_o_ii < 0)[0])
    n_one = len(np.where(f_o_ii > 1)[0])
    assert n_neg == 0
    assert n_one == 0



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module is used to compute the wind energetics using the method
outlined in Vissapragada et al. (2022).
"""

from __future__ import (division, print_function, absolute_import,
                        unicode_literals)
import numpy as np
import astropy.units as u
import astropy.constants as c
from scipy.integrate import simps, cumtrapz
from scipy.interpolate import interp1d


__all__ = ["calculate_mdot_max", "calculate_epsilon_max", "spec_av_cross",
           "compute_column_densities", "compute_transmission_coefficient",
           "calculate_roche_radius", "calculate_f_xuv", "h_photo_cross",
           "heplus_photo_cross", "helium_photo_cross"]

threshes = {'hydrogen': 13.6 * u.eV, 'helium': 24.58 * u.eV,
            'helium+': 54.4 * u.eV}


def calculate_mdot_max(R_pl, M_pl, M_star, a, r_grid, spectrum, n_h, n_he, 
                       n_he_plus):
    """
    Calculates the maximum mass-loss rate of an outflow given the profiles of
    neutral hydrogen, neutral helium, and ionized helium by Equation (20) of
    Vissapragada et al. (2022).

    Parameters
    ----------
    R_pl : ``astropy.Quantity``
        The planetary radius. An astropy unit (like u.Rjup) must be specified.

    M_pl : ``astropy.Quantity``
        The planetary mass. An astropy unit (like u.Mjup) must be specified.

    M_star : ``astropy.Quantity``
        The stellar mass. An astropy unit (like u.Msun) must be specified.

    a : ``astropy.Quantity``
        The semimajor axis of the planetary orbit. An astropy unit (like u.au)
        must be specified.

    r_grid : ``numpy.ndarray``
        The radius grid for the calculation. An astropy unit (like u.Rjup) must
        be specified for each value on the grid.
  
    spectrum : ``dict``
        Spectrum of the host star arriving at the planet covering fluxes up to
        the wavelength corresponding to the energy to ionize hydrogen (13.6 eV,
        or 911.65 Angstrom). Can be generated using ``tools.make_spectrum_dict``
        or ``tools.generate_muscles_spectrum``. Currently we assume that the
        spectrum does not include lower energies than 13.6 eV.

    n_h : ``numpy.ndarray``
        The neutral hydrogen number density profile for the wind. An astropy
        unit (like u.cm**-3) must be specified for each value on the grid.
 
    n_he : ``numpy.ndarray``
        The neutral helium number density profile for the wind. An astropy
        unit (like u.cm**-3) must be specified for each value on the grid.

    n_he_plus : ``numpy.ndarray``
        The (singly-)ionized helium number density profile for the wind. An
        astropy unit (like u.cm**-3) must be specified for each value on the
        grid.

    Returns
    -------
    md : ``astropy.Quantity``
        Maximum mass-loss rate (in g/s) of the wind assuming it is driven only
        by photoionization.
    """
    R_roche = calculate_roche_radius(R_pl, M_pl, M_star, a)
    eps = calculate_epsilon_max(r_grid, spectrum, n_h, n_he, n_he_plus, 
        R_pl, R_roche)
    K = 1 - 1.5 * (R_pl / R_roche) + 0.5 * (R_pl / R_roche) ** 3
    F_XUV = calculate_f_xuv(spectrum)
    md = 4 * eps * np.pi * R_pl.to(u.cm) ** 3 * F_XUV / (K * c.G * M_pl.to(u.g))
    return md.to(u.g/u.s)


def calculate_epsilon_max(r_grid, spectrum, n_h, n_he, n_he_plus, R_pl,
                          R_roche):
    """
    Calculates the maximum mass-loss efficiency of an outflow given the
    profiles of neutral hydrogen, neutral helium, and ionized helium by
    Equation (19) of Vissapragada et al. (2022).

    Parameters
    ----------
    r_grid : ``numpy.ndarray``
        The radius grid for the calculation. An astropy unit (like u.Rjup) must
        be specified for each value on the grid.
  
    spectrum : ``dict``
        Spectrum of the host star arriving at the planet covering fluxes up to
        the wavelength corresponding to the energy to ionize hydrogen (13.6 eV,
        or 911.65 Angstrom). Can be generated using ``tools.make_spectrum_dict``
        or ``tools.generate_muscles_spectrum``. Currently we assume that the
        spectrum does not include lower energies than 13.6 eV.

    n_h : ``numpy.ndarray``
        The neutral hydrogen number density profile for the wind. An astropy
        unit (like u.cm**-3) must be specified for each value on the grid.
 
    n_he : ``numpy.ndarray``
        The neutral helium number density profile for the wind. An astropy
        unit (like u.cm**-3) must be specified for each value on the grid.

    n_he_plus : ``numpy.ndarray``
        The (singly-)ionized helium number density profile for the wind. An
        astropy unit (like u.cm**-3) must be specified for each value on the
        grid.

    R_pl : ``astropy.Quantity``
        The planetary radius. An astropy unit (like u.Rjup) must be specified.

    R_roche : ``astropy.Quantity``
        The Roche radius. An astropy unit (like u.Rjup) must be specified.

    Returns
    -------
    eps : ``float``
        Maximum mass-loss efficiency of the wind assuming it is driven only
        by photoionization.
    """
    N_h, N_he, N_he_plus = compute_column_densities(r_grid, n_h, n_he, 
        n_he_plus)
    t_coef = compute_transmission_coefficient(spectrum, r_grid,
        N_h, N_he, N_he_plus)

    h_av_cross = spec_av_cross(r_grid, spectrum, t_coef, 'hydrogen')
    he_av_cross = spec_av_cross(r_grid, spectrum, t_coef, 'helium')
    he_plus_av_cross = spec_av_cross(r_grid, spectrum, t_coef, 'helium+')

    h_heating_coef = h_av_cross * n_h
    he_heating_coef = he_av_cross * n_he
    he_plus_heating_coef = he_plus_av_cross * n_he_plus
    total_heating_coef = h_heating_coef + he_heating_coef + he_plus_heating_coef
    
    r = r_grid.to(u.cm)
    integrand = total_heating_coef * r ** 2
    mask = (r >= R_pl) & (r <= R_roche)
    eps = simps(integrand[mask], x=r[mask]) * u.cm ** 2 / (R_pl.to(u.cm) ** 2)
    return eps


def spec_av_cross(r_grid, spectrum, t_coef, species):
    """
    Calculates the heating cross-section for photoionization using Equation (16)
    of Vissapragada et al. (2022).

    Parameters
    ----------
    r_grid : ``numpy.ndarray``
        The radius grid for the calculation. An astropy unit (like u.Rjup) must
        be specified for each value on the grid.
  
    spectrum : ``dict``
        Spectrum of the host star arriving at the planet covering fluxes up to
        the wavelength corresponding to the energy to ionize hydrogen (13.6 eV,
        or 911.65 Angstrom). Can be generated using ``tools.make_spectrum_dict``
        or ``tools.generate_muscles_spectrum``. Currently we assume that the
        spectrum does not include lower energies than 13.6 eV.

    t_coef : ``numpy.ndarray``
        The transmission coefficient profile for the wind as a function of 
        frequency and altitude. In the optically-thin part of the outflow this
        should be very close to 1.
 
    species : ``str``
        The photoionzation target for which we are calculating the heating
        cross-section. Must be one of 'hydrogen', 'helium', or 'helium+'.

    Returns
    -------
    cross : ``astropy.Quantity``
        Heating cross-section in cm ** 2 for the selected species.
    """
    wav_grid = spectrum['wavelength'] * spectrum['wavelength_unit']
    flux_grid = spectrum['flux_lambda'] * spectrum['flux_unit']
    flux_grid = flux_grid.to(u.erg / u.s / u.cm / u.cm / u.Hz,
                             equivalencies=u.spectral_density(wav_grid))
    wavs_hz = wav_grid.to(u.Hz, equivalencies=u.spectral())[::-1]
    flux_grid = flux_grid[::-1]
    xx, yy = np.meshgrid(wavs_hz, r_grid)
    
    threshold = threshes[species]
    crosses = {'hydrogen': h_photo_cross, 'helium': helium_photo_cross,
               'helium+': heplus_photo_cross}
    cross = crosses[species]
    evgrid = xx.to(u.eV, equivalencies=u.spectral())
    eta_grid = 1 - threshold/evgrid
    spec_grid, __ = np.meshgrid(flux_grid, r_grid)
    crossgrid = cross(xx)
    crossgrid[xx.to(u.eV, equivalencies=u.spectral()) < threshold] = \
        0. * u.cm ** 2

    numgrid = eta_grid * spec_grid * crossgrid * t_coef
    numgrid = numgrid.to(u.erg / u.s / u.Hz)
    num = simps(numgrid, x=wavs_hz, axis=-1) * u.erg/u.s

    F_XUV = calculate_f_xuv(spectrum)
    cross = num / F_XUV
    return cross.to(u.cm**2)


def compute_column_densities(r_grid, n_h, n_he, n_he_plus):
    """
    Given the density profiles of H, He, and He+, this function calculates the
    column densities.

    Parameters
    ----------
    r_grid : ``numpy.ndarray``
        The radius grid for the calculation. An astropy unit (like u.Rjup) must
        be specified for each value on the grid.

    n_h : ``numpy.ndarray``
        The neutral hydrogen number density profile for the wind. An astropy
        unit (like u.cm ** -3) must be specified for each value on the grid.
 
    n_he : ``numpy.ndarray``
        The neutral helium number density profile for the wind. An astropy
        unit (like u.cm ** -3) must be specified for each value on the grid.

    n_he_plus : ``numpy.ndarray``
        The (singly-)ionized helium number density profile for the wind. An
        astropy unit (like u.cm ** -3) must be specified for each value on the
        grid.

    Returns
    -------
    column_h : ``numpy.ndarray``
        The neutral hydrogen column density profile in u.cm ** -2.

    column_he : ``numpy.ndarray``
        The neutral helium column density profile in u.cm ** -2.

    column_he_plus : ``numpy.ndarray``
        The ionized helium column density profile in u.cm ** -2.
    """
    # Flip grid to integrate from infty to r
    r_grid_temp = r_grid.to(u.cm).value[::-1]
    n_h_temp = n_h[::-1]
    n_he_temp = n_he[::-1]
    n_he_plus_temp = n_he_plus[::-1]
    
    column_h = cumtrapz(n_h_temp, r_grid_temp, initial=0) * u.cm ** -2
    column_he = cumtrapz(n_he_temp, r_grid_temp, initial=0) * u.cm ** -2
    column_he_plus = cumtrapz(n_he_plus_temp, r_grid_temp,
                              initial=0) * u.cm ** -2
    
    # Flip back
    column_h = -column_h[::-1]
    column_he = -column_he[::-1]
    column_he_plus = -column_he_plus[::-1]

    return column_h, column_he, column_he_plus


def compute_transmission_coefficient(spectrum, r_grid, N_h, N_he, N_he_plus):
    """
    Given the column density profiles of H, He, and He+, this function 
    calculates the transmission coefficient (negative exponential of the optical
    depth) as a function of altitude and frequency. 

    Parameters
    ----------
    r_grid : ``numpy.ndarray``
        The radius grid for the calculation. An astropy unit (like u.Rjup) must
        be specified for each value on the grid.

    spectrum : ``dict``
        Spectrum of the host star arriving at the planet covering fluxes up to
        the wavelength corresponding to the energy to ionize hydrogen (13.6 eV,
        or 911.65 Angstrom). Can be generated using ``tools.make_spectrum_dict``
        or ``tools.generate_muscles_spectrum``. Currently we assume that the
        spectrum does not include lower energies than 13.6 eV.

    N_h : ``numpy.ndarray``
        The neutral hydrogen column density profile. An astropy unit (like
        u.cm ** -2) must be specified for each value on the grid.

    N_he : ``numpy.ndarray``
        The neutral helium column density profile. An astropy unit (like
        u.cm ** -2) must be specified for each value on the grid.

    N_he_plus : ``numpy.ndarray``
        The ionized helium column density profile. An astropy unit (like
        u.cm ** -2) must be specified for each value on the grid.

    Returns
    -------
    t_coef : ``numpy.ndarray``
        The transmission coefficient profile for the wind as a function of 
        frequency and altitude.
    """
    wavs = spectrum['wavelength'] * spectrum['wavelength_unit']
    wavs_hz = wavs.to(u.Hz, equivalencies=u.spectral())[::-1]

    cd_h = interp1d(r_grid.to(u.cm).value, N_h.to(u.cm ** -2).value,
                    kind='cubic')
    cd_he = interp1d(r_grid.to(u.cm).value, N_he.to(u.cm ** -2).value,
                     kind='cubic')
    cd_he_plus = interp1d(r_grid.to(u.cm).value, N_he_plus.to(u.cm**-2).value,
                          kind='cubic')

    xx, yy = np.meshgrid(wavs_hz, r_grid)    
    h_cross = h_photo_cross(xx)
    he_cross = helium_photo_cross(xx) 
    he_plus_cross = heplus_photo_cross(xx)

    h_cross[xx.to(u.eV, equivalencies=u.spectral()) < threshes['hydrogen']] = \
        0. * u.cm ** 2
    he_cross[xx.to(u.eV, equivalencies=u.spectral()) < threshes['helium']] = \
        0. * u.cm ** 2
    he_plus_cross[xx.to(u.eV, equivalencies=u.spectral()) <
                  threshes['helium+']] = 0. * u.cm ** 2

    h_column = cd_h(yy.to(u.cm).value) * u.cm ** -2
    he_column = cd_he(yy.to(u.cm).value) * u.cm ** -2
    he_plus_column = cd_he_plus(yy.to(u.cm).value) * u.cm ** -2

    tau_h = h_cross * h_column
    tau_he = he_cross * he_column
    tau_he_plus = he_plus_cross * he_plus_column

    tau_tot = tau_h + tau_he + tau_he_plus
    transmission_coef = np.exp(-tau_tot)
    
    return transmission_coef


def calculate_roche_radius(R_pl, M_pl, M_star, a):
    """
    Calculates the Roche radius for the planet.

    Parameters
    ----------
    R_pl : ``astropy.Quantity``
        The planetary radius. An astropy unit (like u.Rjup) must be specified.

    M_pl : ``astropy.Quantity``
        The planetary mass. An astropy unit (like u.Mjup) must be specified.

    M_star : ``astropy.Quantity``
        The stellar mass. An astropy unit (like u.Msun) must be specified.

    a : ``astropy.Quantity``
        The semimajor axis of the planetary orbit. An astropy unit (like u.au)
        must be specified.
 
    Returns
    -------
    R_roche : ``astropy.Quantity``
        The Roche radius.
    """
    return (a * (M_pl / (3 * M_star)) ** (1 / 3)).to(u.Rjup)


def calculate_f_xuv(spectrum):
    """
    Calculates the total XUV flux given the spectrum at the planet (Equation
    14 of Vissapragada et al. 2022, where the minimum threshold is 13.6 eV. This
    function currently assumes the spectrum is truncated at 13.6 eV and does not
    include lower energies. 

    Parameters
    ----------
    spectrum : ``dict``
        Spectrum of the host star arriving at the planet covering fluxes up to
        the wavelength corresponding to the energy to ionize hydrogen (13.6 eV,
        or 911.65 Angstrom). Can be generated using ``tools.make_spectrum_dict``
        or ``tools.generate_muscles_spectrum``. Currently we assume that the
        spectrum does not include lower energies than 13.6 eV.

    Returns
    -------
    f_xuv : ``astropy.Quantity``
        The integrated XUV flux.
    """
    wav_grid = spectrum['wavelength'] * spectrum['wavelength_unit']
    flux_grid = spectrum['flux_lambda'] * spectrum['flux_unit']
    flux_grid = flux_grid.to(u.erg / u.s / u.cm / u.cm / u.Hz,
                             equivalencies=u.spectral_density(wav_grid))
    wavs_hz = wav_grid.to(u.Hz, equivalencies=u.spectral())[::-1]
    flux_grid = flux_grid[::-1]
    f_xuv = simps(flux_grid, x=wavs_hz) * u.erg / u.s / u.cm ** 2
    return f_xuv


def h_photo_cross(nu_init):
    """
    Calculates the photoionization cross-section of hydrogen (Equation 17 from
    Vissapragada et al. 2022). Note: potential overlap with 
    ``microphysics.hydrogen_cross_section``.

    Parameters
    ----------
    nu_init : ``numpy.ndarray``
        Frequencies at which to compute the cross-section. An astropy unit
        (like u.Hz) must be specified for each value in the array.

    Returns
    -------
    out : ``numpy.ndarray``
        The cross-section in cm ** 2.
    """
    "nu in Hz -> cross section in units cm^2"
    threshold_energy = threshes['hydrogen']
    threshold_cross = 6.3e-18 * u.cm * u.cm
    
    nu = nu_init.to(u.eV, equivalencies=u.spectral())
    
    eps = np.sqrt(nu / threshold_energy - 1)
    arg1 = threshold_cross*np.exp(4 - 4 * np.arctan(eps) / eps / u.rad)
    arg2 = (1 - np.exp(-2 * np.pi / eps))
    arg3 = (threshold_energy / nu) **4
    
    out = arg1 / arg2 * arg3
    return out


def heplus_photo_cross(nu_init):
    """
    Calculates the photoionization cross-section of ionized helium (Equation 17
    from Vissapragada et al. 2022).

    Parameters
    ----------
    nu_init : ``numpy.ndarray``
        Frequencies at which to compute the cross-section. An astropy unit
        (like u.Hz) must be specified for each value in the array.

    Returns
    -------
    out : ``numpy.ndarray``
        The cross-section in cm**2. 
    """
    "nu in Hz -> cross section in units cm^2"
    threshold_energy = threshes['helium+']
    threshold_cross = 6.3e-18 / 4 * u.cm * u.cm
    
    nu = nu_init.to(u.eV, equivalencies=u.spectral())
    
    eps = np.sqrt(nu/threshold_energy - 1)
    arg1 = threshold_cross * np.exp(4 - 4 * np.arctan(eps) / eps / u.rad)
    arg2 = (1 - np.exp(-2 * np.pi / eps))
    arg3 = (threshold_energy / nu) **4
    
    out = arg1 / arg2 * arg3
    return out


def helium_photo_cross(nu_init):
    """
    Calculates the photoionization cross-section of neutral helium. This is 
    Equation 18 from Vissapragada et al. 2022, which itself is from Yan, 
    Sadeghpour, & Dalgarno (1998, ApJ). Note that this does not overlap with
    the state-resolved cross-sections from the ``microphysics`` module as this
    is a total photoionization cross-section across all states. 

    Parameters
    ----------
    nu_init : ``numpy.ndarray``
        Frequencies at which to compute the cross-section. An astropy unit
        (like u.Hz) must be specified for each value in the array.

    Returns
    -------
    out : ``numpy.ndarray``
        The cross-section in cm**2. 
    """
    # nu in Hz -> cross section in units cm^2
    # Source: Yan, Sadeghpour, & Dalgarno (1998, ApJ)
    threshold_energy = threshes['helium']
    
    nu = nu_init.to(u.eV, equivalencies=u.spectral())
    
    nu_masked = nu
    x = nu_masked / threshold_energy
    e_term = (nu_masked / u.eV / 1000.)**(7 / 2)
    
    coefs = [-4.7416, 14.8200, -30.8678, 37.3584, -23.4585, 5.9133]
    sum_term = 0.
    for i, coef in enumerate(coefs):
        sum_term += coef / x ** ((i + 1) / 2)
    
    num = 733 * u.barn
    out = num.to(u.cm * u.cm)/e_term * (1 + sum_term)

    return out



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import astropy.units as u
from p_winds import hydrogen, tools


# HD 209458 b
R_pl = (1.39 * u.jupiterRad).value
M_pl = (0.73 * u.jupiterMass).value
m_dot = (8E10 * u.g / u.s).value
T_0 = (9E3 * u.K).value
h_fraction = 0.9
average_f_ion = 0.7
data_test_url = 'https://raw.githubusercontent.com/ladsantos/p-winds/main/data/solar_spectrum_scaled_lambda.dat'

r = np.linspace(1, 15, 500)


# Let's test if the code is producing reasonable outputs. The ``ion_fraction()``
# function for HD 209458 b should produce a profile with an ion fraction of
# approximately one near the planetary surface, and approximately 4E-4 in the
# outer layers.
def test_ion_fraction_spectrum(precision_threshold=1E-4):
    units = {'wavelength': u.angstrom, 'flux': u.erg / u.s / u.cm ** 2 /
                                               u.angstrom}
    spectrum = tools.make_spectrum_from_file(data_test_url, units)

    # Test the approximate photoionization
    f_r = hydrogen.ion_fraction(r, R_pl, T_0, h_fraction, m_dot, M_pl,
                                average_f_ion,
                                spectrum_at_planet=spectrum,
                                relax_solution=True)
    assert abs((f_r[-1] - 0.998737) / f_r[-1]) < precision_threshold

    # Test the exact photoionization
    f_r = hydrogen.ion_fraction(r, R_pl, T_0, h_fraction, m_dot, M_pl,
                                average_f_ion,
                                spectrum_at_planet=spectrum,
                                relax_solution=True, exact_phi=True)
    assert abs((f_r[-1] - 0.998913) / f_r[-1]) < precision_threshold



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module computes the neutral and ionized populations of O in the upper
atmosphere.
"""

from __future__ import (division, print_function, absolute_import,
                        unicode_literals)
import numpy as np
import astropy.units as u
import astropy.constants as c
from scipy.integrate import simps, solve_ivp, odeint
from p_winds import tools, microphysics
import warnings


__all__ = ["radiative_processes", "electron_impact_ionization", "recombination",
           "charge_transfer", "ion_fraction"]


# Some hard coding based on the astrophysical literature
_SOLAR_OXYGEN_ABUNDANCE_ = 8.69  # Asplund et al. 2009
_SOLAR_OXYGEN_FRACTION_ = 10 ** (_SOLAR_OXYGEN_ABUNDANCE_ - 12.00)


# Photoionization of O I (neutral)
def radiative_processes(spectrum_at_planet):
    """
    Calculate the photoionization rate of oxygen at null optical depth based
    on the EUV spectrum arriving at the planet.

    Parameters
    ----------
    spectrum_at_planet : ``dict``
        Spectrum of the host star arriving at the planet covering fluxes at
        least up to the wavelength corresponding to the energy to ionize
        oxygen (13.62 eV, or 910 Angstrom).

    Returns
    -------
    phi_oi : ``float``
        Ionization rate of O I at null optical depth in unit of 1 / s.

    a_oi : ``float``
        Flux-averaged photoionization cross-section of O I in unit of cm ** 2.

    a_h_oi : ``float``
        Flux-averaged photoionization cross-section of H I in the range absorbed
        by O I in unit of cm ** 2.

    a_he : ``float``
        Flux-averaged photoionization cross-section of He I in unit of cm ** 2.
    """
    wavelength = (spectrum_at_planet['wavelength'] *
                  spectrum_at_planet['wavelength_unit']).to(u.angstrom).value
    flux_lambda = (spectrum_at_planet['flux_lambda'] * spectrum_at_planet[
        'flux_unit']).to(u.erg / u.s / u.cm ** 2 / u.angstrom).value
    energy = ((c.h * (c.c / wavelength / u.angstrom).to(u.Hz)).to(u.eV)).value
    energy_erg = (energy * u.eV).to(u.erg).value

    # Auxiliary definitions
    parameters_dict = microphysics.sigma_properties_v1996()

    energy_threshold_oi = parameters_dict['O I'][0]  # Ionization threshold in
    # eV
    wl_break_oi = 12398.42 / energy_threshold_oi  # O I ionization threshold in
    # angstrom
    wl_break_he = 504  # He ionization threshold in angstrom
    i0 = tools.nearest_index(wavelength, wl_break_he)
    i1 = tools.nearest_index(wavelength, wl_break_oi)
    wavelength_cut_0 = wavelength[:i0]
    flux_lambda_cut_0 = flux_lambda[:i0]
    wavelength_cut_1 = wavelength[:i1]
    flux_lambda_cut_1 = flux_lambda[:i1]
    energy_cut_1 = energy_erg[:i1]

    # Calculate the photoionization cross-section
    a_lambda_oi = microphysics.general_cross_section(wavelength_cut_1,
                                                     species='O I')

    # The flux-averaged photoionization cross-section of O I
    a_oi = abs(simps(flux_lambda_cut_1 * a_lambda_oi, wavelength_cut_1) /
               simps(flux_lambda_cut_1, wavelength_cut_1))

    # The flux-averaged photoionization cross-section of H is also going to be
    # needed because it adds to the optical depth that O I see.
    a_lambda_h_oi = microphysics.hydrogen_cross_section(
        wavelength=wavelength_cut_1)
    a_h_oi = abs(simps(flux_lambda_cut_1 * a_lambda_h_oi, wavelength_cut_1) /
                 simps(flux_lambda_cut_1, wavelength_cut_1))

    # Same for the He atoms, but only up to the He ionization threshold
    a_lambda_he = microphysics.helium_total_cross_section(wavelength_cut_0)
    a_he = abs(simps(flux_lambda_cut_0 * a_lambda_he, wavelength_cut_0) /
               simps(flux_lambda_cut_0, wavelength_cut_0))

    # Calculate the photoionization rates
    phi_oi = abs(simps(flux_lambda_cut_1 * a_lambda_oi / energy_cut_1,
                       wavelength_cut_1))

    return phi_oi, a_oi, a_h_oi, a_he


# Ionization rate of O by electron impact
def electron_impact_ionization(electron_temperature):
    """
    Calculates the electron impact ionization rate that consumes neutral O and
    produces singly-ionized O. Based on the formula of Voronov 1997
    (https://ui.adsabs.harvard.edu/abs/1997ADNDT..65....1V/abstract).

    Parameters
    ----------
    electron_temperature : ``float``
        Temperature of the plasma where the electrons are embedded in unit of
        Kelvin.

    Returns
    -------
    ionization_rate_oi : ``float``
        Ionization rate of neutral O into singly-ionized O in unit of
        cm ** 3 / s.
    """
    boltzmann_constant = 8.617333262145179e-05  # eV / K
    electron_energy = boltzmann_constant * electron_temperature
    energy_ratio_oi = 11.3 / electron_energy
    ionization_rate_oi = 3.59E-8 * (0.073 + energy_ratio_oi) ** (-1) * \
        energy_ratio_oi ** 0.34 * np.exp(-energy_ratio_oi)
    return ionization_rate_oi


# Recombination of singly-ionized O into neutral O
def recombination(electron_temperature):
    """
    Calculates the rate of recombination of singly-ionized O with an electron to
    produce a neutral O atom. Based on the formulation of Woodall et al. 2007
    (https://ui.adsabs.harvard.edu/abs/2007A%26A...466.1197W/abstract).

    Parameters
    ----------
    electron_temperature : ``float``
        Temperature of the plasma where the electrons are embedded in unit of
        Kelvin.

    Returns
    -------
    alpha_rec_oi  : ``float``
        Recombination rate of O II into O I in units of cm ** 3 / s.
    """
    alpha_rec_oi = 3.25E-12 * (300 / electron_temperature) ** 0.66
    return alpha_rec_oi


# Charge transfer between O and H
def charge_transfer(temperature):
    """
    Calculates the charge exchange rates of O with H nuclei. Based on the
    formulation of Woodall et al. 2007
    (https://ui.adsabs.harvard.edu/abs/2007A%26A...466.1197W/abstract).

    Parameters
    ----------
    temperature : ``float``
        Isothermal temperature of the upper atmosphere in unit of Kelvin.

    Returns
    -------
    ct_rate_oi_hp : ``float``
        Charge transfer rate between neutral O and H+ in units of cm ** 3 / s.

    ct_rate_oii_h : ``float``
        Charge transfer rate between O+ and neutral H in units of cm ** 3 / s.
    """
    # Recombination of O II into O I
    ct_rate_oii_h = 5.66E-10 * (300 / temperature) ** (-0.36) * \
        np.exp(8.6 / temperature)

    # Ionization of O I into O II
    ct_rate_oi_hp = 7.31E-10 * (300 / temperature) ** (-0.23) * \
        np.exp(-226 / temperature)

    return ct_rate_oi_hp, ct_rate_oii_h


# Calculation the number fractions of O II
def ion_fraction(radius_profile, velocity, density, hydrogen_ion_fraction,
                 helium_ion_fraction, planet_radius, temperature, h_fraction,
                 speed_sonic_point, radius_sonic_point, density_sonic_point,
                 spectrum_at_planet, o_fraction=_SOLAR_OXYGEN_FRACTION_,
                 initial_f_o_ion=0.0, relax_solution=False, convergence=0.01,
                 max_n_relax=10, method='Radau', return_rates=False,
                 **options_solve_ivp):
    """
    Calculate the fraction of ionized oxygen in the upper atmosphere in function
    of the radius in unit of planetary radius.

    Parameters
    ----------
    radius_profile : ``numpy.ndarray``
        Radius in unit of planetary radii.

    velocity : ``numpy.ndarray``
         Velocities sampled at the values of ``radius_profile`` in units of
         sound speed. Similar to the output of ``parker.structure()``.

    density : ``numpy.ndarray``
        Densities sampled at the values of ``radius_profile`` in units of
        density at the sonic point. Similar to the output of
        ``parker.structure()``.

    hydrogen_ion_fraction : ``numpy.ndarray``
        Number fraction of H ion over total H in the upper atmosphere in
        function of radius. Similar to the output of
        ``hydrogen.ion_fraction()``.

    helium_ion_fraction : ``numpy.ndarray``
        Number fraction of He ion over total He in the upper atmosphere in
        function of radius. Similar to the output of
        ``helium.population_fraction()``, but should be ``1 - f_1_r - f_3_r``.

    planet_radius : ``float``
        Planetary radius in unit of Jupiter radius.

    temperature : ``float``
        Isothermal temperature of the upper atmosphere in unit of Kelvin.

    h_fraction : ``float``
        Total (ion + neutral) H number fraction of the atmosphere.

    speed_sonic_point : ``float``
        Speed of sound in the outflow in units of km / s.

    radius_sonic_point : ``float``
        Radius of the sonic point in unit of Jupiter radius.

    density_sonic_point : ``float``
        Density at the sonic point in units of g / cm ** 3.

    spectrum_at_planet : ``dict``
        Spectrum of the host star arriving at the planet covering fluxes at
        least up to the wavelength corresponding to the energy to ionize
        oxygen (13.62 eV, or 910 Angstrom). Can be generated using
        ``tools.make_spectrum_dict``.

    o_fraction : ``float``, optional
        Fraction of total oxygen in the upper atmosphere. Default value assumes
        solar abundance.

    initial_f_o_ion : ``float``, optional
        The initial oxygen ion fraction at the layer near the surface of the
        planet. Default is 0.0, i.e., 100% neutral.

    relax_solution : ``bool``, optional
        The first solution is calculating by initially assuming the entire
        atmosphere is in neutral state. If ``True``, the solution will be
        re-calculated in a loop until it converges to a delta_f of 1%, or for a
        maximum of 10 loops (default parameters). Default is ``False``.

    convergence : ``float``, optional
        Value of delta_f at which to stop the relaxation of the solution for
        ``f_r``. Default is 0.01.

    max_n_relax : ``int``, optional
        Maximum number of loops to perform the relaxation of the solution for
        the ion fractions. Default is 10.

    method : ``str``, optional
        If method is ``'odeint'``, then ``scipy.integrate.odeint()`` is used
        instead of ``scipy.integrate.solve_ivp()`` to calculate the steady-state
        distribution of helium. Any other method will fall back to an option of
        ``solve_ivp()`` methods. For example, if ``method`` is set to
        ``'Radau'``, then use ``solve_ivp(method='Radau')``. Default is
        ``'Radau'``.

    return_rates : ``bool``, optional
        If ``True``, then this function also returns a ``dict`` object
        containing the various reaction rates in function of radius and in units
        of 1 / s. Default is ``False``.

    **options_solve_ivp:
        Options to be passed to the ``scipy.integrate.solve_ivp()`` solver. You
        may want to change the options ``atol`` (absolute tolerance; default is
        1E-6) or ``rtol`` (relative tolerance; default is 1E-3). If you are
        having numerical issues, you may want to decrease the tolerance by a
        factor of 10 or 100, or 1000 in extreme cases.

    Returns
    -------
    f_oii_r : ``numpy.ndarray``
            Fraction of singly-ionized oxygen in function of radius.

    reaction_rates : ``dict``
        Dictionary containing the reaction rates in function of radius and in
        units of 1 / s. Only returned when ``return_rates`` is set to ``True``.
        Here is a short description of the dict keys:

        * `photoionization`: Photoionization of O I into O II
        * `recombination`: Recombination of O II into O I
        * `e_impact_ionization`: Electron impact ionization of O I into O II
        * `charge_exchange_HII`: Charge exchange between O I and H II
        * `charge_exchange_HI`: Charge exchange between O II and H I
    """
    vs = speed_sonic_point  # km / s
    rs = radius_sonic_point  # jupiterRad
    rhos = density_sonic_point  # g / cm ** 3

    # Recombination rates of C in unit of rs ** 2 * vs
    alpha_rec_unit = ((rs * 7.1492E+09) ** 2 * vs * 1E5)  # cm ** 3 / s
    alpha_rec_oi = recombination(temperature)
    alpha_rec_oi = alpha_rec_oi / alpha_rec_unit

    # Hydrogen mass in unit of rhos * rs ** 3
    m_h_unit = (rhos * (rs * 7.1492E+09) ** 3)  # Converted to g
    m_h = 1.67262192E-24 / m_h_unit

    # Photoionization rates at null optical depth at the distance of the planet
    # from the host star, in unit of vs / rs, and the flux-averaged
    # cross-sections in units of rs ** 2
    phi_unit = vs * 1E5 / rs / 7.1492E+09  # 1 / s
    phi_oi, a_oi, a_h_oi, a_he = radiative_processes(spectrum_at_planet)
    phi_oi = phi_oi / phi_unit
    a_oi = a_oi / (rs * 7.1492E+09) ** 2
    a_h_oi = a_h_oi / (rs * 7.1492E+09) ** 2
    a_he = a_he / (rs * 7.1492E+09) ** 2

    # Electron-impact ionization rate for C I in the same unit as the
    # recombination rates
    ionization_rate_oi = electron_impact_ionization(temperature)
    ionization_rate_oi = ionization_rate_oi / alpha_rec_unit

    # Charge transfer rates in the same unit as the recombination rates
    ct_rate_oi_hp, ct_rate_oii_h = charge_transfer(temperature)
    ct_rate_oii_h = ct_rate_oii_h / alpha_rec_unit
    ct_rate_oi_hp = ct_rate_oi_hp / alpha_rec_unit

    # We solve the steady-state ionization balance in a similar way that we do
    # for He

    # The radius in unit of radius at the sonic point
    r = radius_profile * planet_radius / rs
    dr = np.diff(r)
    dr = np.concatenate((dr, np.array([dr[-1], ])))

    # With all this setup done, now we need to assume something about the
    # distribution of neutral O in the atmosphere. Let's assume it based on the
    # initial guess input.
    column_density = np.flip(np.cumsum(np.flip(dr * density)))  # Total column
    # density
    column_density_h_0 = np.flip(  # Column density of atomic H only
        np.cumsum(np.flip(dr * density * (1 - hydrogen_ion_fraction))))
    he_fraction = 1 - h_fraction
    column_density_he_0 = np.flip(  # Column density of atomic He only
        np.cumsum(np.flip(dr * density * he_fraction *
                          (1 - helium_ion_fraction))))
    k1 = h_fraction / (h_fraction + 4 * he_fraction + 16 * o_fraction) / m_h
    k2 = he_fraction / (h_fraction + 4 * he_fraction + 16 * o_fraction) / m_h
    k3 = o_fraction / (h_fraction + 4 * he_fraction + 16 * o_fraction) / m_h
    tau_oi_h = k1 * a_h_oi * column_density_h_0
    tau_o_he = k2 * a_he * column_density_he_0
    tau_oi = (1 - initial_f_o_ion) * k3 * a_oi * column_density + tau_oi_h + \
        tau_o_he

    # The differential equation
    def _fun(_r, y, rates=False):
        f_oii = y

        _v = np.interp(_r, r, velocity)
        _rho = np.interp(_r, r, density)
        f_h_ion = np.interp(_r, r, hydrogen_ion_fraction)  # Fraction of H+
        f_he_ion = np.interp(_r, r, helium_ion_fraction)  # Fraction of He+

        # Assume the number density of electrons is equal to the number density
        # of H ions + He ions
        # Since we may run into loss of numerical precision here (big numbers),
        # we manipulate the equations to avoid this problem. It looks a bit
        # messy, but it is necessary
        log_term_1 = np.log(k1) + np.log(_rho)  # H ions
        log_term_2 = np.log(k2) + np.log(_rho)  # He ions
        ionization_rate_oi_n_e = \
            np.exp(log_term_1 + np.log(ionization_rate_oi)) * f_h_ion + \
            np.exp(log_term_2 + np.log(ionization_rate_oi)) * f_he_ion
        ct_rate_oi_hp_n_h_plus = \
            np.exp(log_term_1 + np.log(ct_rate_oi_hp)) * f_h_ion
        alpha_rec_oi_n_e = \
            np.exp(log_term_1 + np.log(alpha_rec_oi)) * f_h_ion + \
            np.exp(log_term_2 + np.log(alpha_rec_oi)) * f_he_ion
        ct_rate_oii_h_n_h0 = \
            np.exp(log_term_1 + np.log(ct_rate_oii_h)) * (1 - f_h_ion)

        # n_e = k1 * _rho * f_h_ion + k2 * _rho * f_he_ion  # Number density of
        # # electrons
        # n_h_plus = k1 * _rho * f_h_ion    # Number density of ionized H
        # n_h0 = k1 * _rho * (1 - f_h_ion)  # Number density of atomic H

        # Terms of dfoii_dr
        t_oi = np.interp(_r, r, tau_oi)
        term1 = (1 - f_oii) * phi_oi * np.exp(-t_oi)  # Photoionization
        term2 = (1 - f_oii) * ionization_rate_oi_n_e  # Electron-impact
        # ionization
        term3 = (1 - f_oii) * ct_rate_oi_hp_n_h_plus  # Charge exchange with
        # H+
        term4 = f_oii * alpha_rec_oi_n_e  # Recombination of O II into O I
        term5 = f_oii * ct_rate_oii_h_n_h0  # Charge exchange of O II with
        # neutral H
        dfoii_dr = (term1 + term2 + term3 - term4 - term5) / _v

        if rates is False:
            return dfoii_dr
        else:
            return np.array([term1, term4, term2, term3, term5]) * phi_unit

    if method == 'odeint':
        # Since 'odeint' yields only warnings when precision is lost or when
        # there is a problem, we transform these warnings into an exception
        with warnings.catch_warnings():
            warnings.filterwarnings("error")
            try:
                sol = odeint(_fun, y0=initial_f_o_ion, t=r, tfirst=True)
            except Warning:
                raise RuntimeError('The solver ``odeint`` failed to obtain a '
                                   'solution.')
        f_oii_r = np.copy(sol).T[0]
    else:
        # We solve it using `scipy.solve_ivp`
        sol = solve_ivp(_fun, (r[0], r[-1],), np.array([initial_f_o_ion, ]),
                        t_eval=r, method=method, **options_solve_ivp)
        f_oii_r = sol['y'][0]
        # When `solve_ivp` has problems, it may return an array with different
        # size than `r`. So we raise an exception if this happens
        if len(f_oii_r) != len(r):
            raise RuntimeError('The solver ``solve_ivp`` failed to obtain a'
                               ' solution.')

    # High densities can be numerically unstable and produce unphysical values
    # of `f_r`, so we replace negative values with zero and values above 1.0
    # with 1.0
    f_oii_r[f_oii_r < 0] = 1E-15
    f_oii_r[f_oii_r > 1.0] = 1.0

    # For the sake of self-consistency, there is the option of repeating the
    # calculation of f_r by updating the optical depth with the new ion
    # fractions.
    if relax_solution is True:
        for i in range(max_n_relax):
            previous_f_oii_r = np.copy(f_oii_r)

            # Re-calculate the column densities
            tau_oi = \
                k3 * a_oi * np.flip(np.cumsum(
                    np.flip(dr * density * (1 - f_oii_r)))) + tau_oi_h + \
                tau_o_he

            # Solve it again
            if method == 'odeint':
                sol = odeint(_fun, y0=initial_f_o_ion, t=r, tfirst=True)
                f_oii_r = np.copy(sol).T[0]
            else:
                sol = solve_ivp(_fun, (r[0], r[-1],),
                                np.array([initial_f_o_ion, ]), t_eval=r,
                                method=method, **options_solve_ivp)
                f_oii_r = sol['y'][0]

            # Replace negative values with zero and values above 1.0 with
            # 1.0
            f_oii_r[f_oii_r < 0] = 1E-15
            f_oii_r[f_oii_r > 1.0] = 1.0

            # Calculate the relative change of f_ion in the outer shell of
            # the atmosphere (where we expect the most important change)
            relative_delta_f_oii = abs(np.sum(f_oii_r - previous_f_oii_r)) \
                / np.sum(previous_f_oii_r)

            # Break the loop if convergence is achieved
            if relative_delta_f_oii < convergence:
                break
            else:
                pass
    else:
        pass

    if return_rates is False:
        return f_oii_r
    else:
        ionization_rate, recombination_rate, e_impact_ion, ch_exchange_hii,\
            ch_exchange_hi = _fun(r, f_oii_r, rates=True)
        reaction_rates = {'photoionization': ionization_rate,
                          'recombination': recombination_rate,
                          'e_impact_ionization': e_impact_ion,
                          'charge_exchange_HII': ch_exchange_hii,
                          'charge_exchange_HI': ch_exchange_hi}
        return f_oii_r, reaction_rates



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module tabulated information about spectral lines relevant to atmospheric
escape detection.
"""


__all__ = ["he_3_properties", "c_ii_properties", "o_i_properties"]


# Line properties of the 1.083 microns He triplet taken from the NIST database
# https://www.nist.gov/pml/atomic-spectra-database
def he_3_properties():
    """
    Returns the central wavelengths in air, oscillator strengths and the
    Einstein coefficient of the helium triplet in 1.083 microns. The values
    were taken from the NIST database:
    https://www.nist.gov/pml/atomic-spectra-database

    Returns
    -------
    lambda_0 : ``float``
        Central wavelength in air of line 0 in unit of m.

    lambda_1 : ``float``
        Central wavelength in air of line 1 in unit of m.

    lambda_2 : ``float``
        Central wavelength in air of line 2 in unit of m.

    f_0 : ``float``
        Oscillator strength of line 0 (unitless).

    f_1 : ``float``
        Oscillator strength of line 1 (unitless).

    f_2 : ``float``
        Oscillator strength of line 2 (unitless).

    a_ij : ``float``
        Einstein coefficient of the whole triplet in unit of 1 / s.
    """
    # Central wavelengths in units of m
    lambda_0 = 1.082909114 * 1E-6
    lambda_1 = 1.083025010 * 1E-6
    lambda_2 = 1.083033977 * 1E-6
    # Oscillator strengths
    f_0 = 5.9902e-02
    f_1 = 1.7974e-01
    f_2 = 2.9958e-01
    # Einstein coefficient in units of s ** (-1)
    a_ij = 1.0216e+07

    return lambda_0, lambda_1, lambda_2, f_0, f_1, f_2, a_ij


# Line properties of the C II lines available with HST/STIS taken from the NIST
# database https://www.nist.gov/pml/atomic-spectra-database
def c_ii_properties():
    """
    Returns the central wavelengths in air, oscillator strengths and the
    Einstein coefficient of the C II lines in the STIS FUV wavelength range. The
    values were taken from the NIST database:
    https://www.nist.gov/pml/atomic-spectra-database

    Returns
    -------
    lambda_0 : ``float``
        Central wavelength in air of line 0 in unit of m.

    lambda_1 : ``float``
        Central wavelength in air of line 1 in unit of m.

    lambda_2 : ``float``
        Central wavelength in air of line 2 in unit of m.

    f_0 : ``float``
        Oscillator strength of line 0 (unitless).

    f_1 : ``float``
        Oscillator strength of line 1 (unitless).

    f_2 : ``float``
        Oscillator strength of line 2 (unitless).

    a_ij_0 : ``float``
        Einstein coefficient of line 0 in unit of 1 / s.

    a_ij_1 : ``float``
        Einstein coefficient of line 1 in unit of 1 / s.

    a_ij_2 : ``float``
        Einstein coefficient of line 2 in unit of 1 / s.
    """
    # Central wavelengths in units of m
    lambda_0 = 1334.5323 * 1E-10
    lambda_1 = 1335.6628 * 1E-10
    lambda_2 = 1335.7079 * 1E-10
    # Oscillator strengths
    f_0 = 1.29e-01
    f_1 = 1.27e-02
    f_2 = 1.15e-01
    # Einstein coefficient in units of s ** (-1)
    a_ij_0 = 2.41e+08
    a_ij_1 = 4.76e+07
    a_ij_2 = 2.88e+08

    return lambda_0, lambda_1, lambda_2, f_0, f_1, f_2, a_ij_0, a_ij_1, a_ij_2


# Line properties of the O I lines available with HST/STIS taken from the NIST
# database https://www.nist.gov/pml/atomic-spectra-database
def o_i_properties():
    """
    Returns the central wavelengths in air, oscillator strengths and the
    Einstein coefficient of the O I lines in the STIS FUV wavelength range. The
    values were taken from the NIST database:
    https://www.nist.gov/pml/atomic-spectra-database

    Returns
    -------
    lambda_0 : ``float``
        Central wavelength in air of line 0 in unit of m.

    lambda_1 : ``float``
        Central wavelength in air of line 1 in unit of m.

    lambda_2 : ``float``
        Central wavelength in air of line 2 in unit of m.

    f_0 : ``float``
        Oscillator strength of line 0 (unitless).

    f_1 : ``float``
        Oscillator strength of line 1 (unitless).

    f_2 : ``float``
        Oscillator strength of line 2 (unitless).

    a_ij_0 : ``float``
        Einstein coefficient of line 0 in unit of 1 / s.

    a_ij_1 : ``float``
        Einstein coefficient of line 1 in unit of 1 / s.

    a_ij_2 : ``float``
        Einstein coefficient of line 2 in unit of 1 / s.
    """
    # Central wavelengths in units of m
    lambda_0 = 1302.168 * 1E-10
    lambda_1 = 1304.858 * 1E-10
    lambda_2 = 1306.029 * 1E-10
    # Oscillator strengths
    f_0 = 5.20e-02
    f_1 = 5.18e-02
    f_2 = 5.19e-02
    # Einstein coefficient in units of s ** (-1)
    a_ij_0 = 3.41e+08
    a_ij_1 = 2.03e+08
    a_ij_2 = 6.76e+07

    return lambda_0, lambda_1, lambda_2, f_0, f_1, f_2, a_ij_0, a_ij_1, a_ij_2



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module contains some useful hard-coded data and equations for calculations
in the other modules.
"""

from __future__ import (division, print_function, absolute_import,
                        unicode_literals)
import numpy as np
import astropy.units as u
import astropy.constants as c
from warnings import warn, filterwarnings

# Ignore some annoying warnings that are not a problem when using the MUSCLES
# spectra
filterwarnings(action='ignore', category=RuntimeWarning)


__all__ = ["hydrogen_cross_section", "helium_total_cross_section",
           "helium_singlet_cross_section", "helium_triplet_cross_section",
           "he_collisional_strength", "general_cross_section",
           "sigma_properties_v1996"]


# Photoionization cross-section of hydrogen
def hydrogen_cross_section(wavelength=None, energy=None):
    """
    Compute the photoionization cross-section of hydrogen in function of
    wavelength or energy.

    Parameters
    ----------
    wavelength : ``float`` or ``numpy.ndarray``, optional
        Wavelength in unit of angstrom. Default is ``None``. If ``None``,
        ``energy`` cannot be ``None``.

    energy : ``float`` or ``numpy.ndarray``, optional
        Energy in unit of electron-volt. Default is ``None``. If ``None``,
        ``wavelength`` cannot be ``None``.

    Returns
    -------
    a_lambda : ``float`` or ``numpy.ndarray``
        Cross-section in function of wavelength and in unit of cm ** 2. Only
        returned if wavelength was input.

    a_nu : ``float`` or ``numpy.ndarray``
        Cross-section in function of energy and in unit of cm ** 2. Only
        returned if energy was input.
    """

    if wavelength is not None:
        epsilon = (911.65 / wavelength - 1) ** 0.5
        # Photoionization cross-section in function of wavelength
        a_lambda = \
            (6.3E-18 * np.exp(4 - (4 * np.arctan(epsilon)) / epsilon) /
             (1 - np.exp(-2 * np.pi / epsilon)) * (wavelength / 911.65)
             ** 4)
        if isinstance(a_lambda, np.ndarray):
            a_lambda[np.isnan(a_lambda)] = 0.0  # Remove NaNs if necessary
        return a_lambda

    elif energy is not None:
        epsilon = (energy / 13.6 - 1) ** 0.5
        # Photoionization cross-section in function of energy
        a_nu = (6.3E-18 * np.exp(4 - (4 * np.arctan(epsilon)) / epsilon) /
                (1 - np.exp(-2 * np.pi / epsilon)) * (13.6 / energy) ** 4)
        if isinstance(a_nu, np.ndarray):
            a_nu[np.isnan(a_nu)] = 0.0  # Remove NaNs if necessary
        return a_nu

    else:
        raise ValueError('Either the wavelength or energy has to be provided.')


# Total He photoionization cross-section
def helium_total_cross_section(wavelength):
    """
    Compute the total photoionization cross-section of helium in function of
    wavelength.

    Source: Yan, Sadeghpour, & Dalgarno (1998, ApJ)

    Parameters
    ----------
    wavelength : ``numpy.ndarray``
        Wavelength in unit of angstrom.

    Returns
    -------
    a_lambda_1 : ``numpy.ndarray``
        Cross-section in function of wavelength and in unit of cm ** 2.
    """
    nu_init = (wavelength * u.AA).to(u.Hz, equivalencies=u.spectral())
    threshold_energy = 24.58 * u.eV
    nu = nu_init.to(u.eV, equivalencies=u.spectral())
    x = nu / threshold_energy
    e_term = (nu / u.eV / 1000.) ** (7 / 2)
    
    coefs = [-4.7416, 14.8200, -30.8678, 37.3584, -23.4585, 5.9133]
    sum_term = 0.
    for i, coef in enumerate(coefs):
        sum_term += coef / x**((i+1)/2)
    
    num = 733 * u.barn
    out = num.to(u.cm * u.cm) / e_term * (1 + sum_term)
    out[nu < threshold_energy] = 0. * (u.cm ** 2)

    return out.to(u.cm ** 2).value


# Photoionization cross-section of helium singlet
def helium_singlet_cross_section(wavelength):
    """
    Compute the photoionization cross-section of helium singlet in function of
    wavelength.

    Parameters
    ----------
    wavelength : ``float`` or ``numpy.ndarray``
        Wavelength in unit of angstrom.

    Returns
    -------
    a_lambda_1 : ``float`` or ``numpy.ndarray``
        Cross-section in function of wavelength and in unit of cm ** 2.
    """
    energy = 12398.41984332 / wavelength  # Energy in unit of eV

    # The cross-sections for helium are partially hard-coded and based on those
    # of hydrogen
    a_lambda_h = hydrogen_cross_section(wavelength=wavelength)

    # For the singlet, we simply scale the cross-sections of H following the
    # results from Brown (1971; ADS:1971ApJ...164..387B)
    scale = np.ones_like(wavelength)
    scale *= 37.0 - 19.1 * (energy / 65.4) ** (-0.76)
    # Clip negative values of scale
    scale[scale < 0] = 0.0
    a_lambda_1 = a_lambda_h * scale

    return a_lambda_1


# Photoionization cross-section of helium triplet
def helium_triplet_cross_section():
    """
    Compute the photoionization cross-section of helium triplet in function of
    wavelength.

    Returns
    -------
    wavelength : ``numpy.ndarray``
        Wavelength in which the cross-section was sampled in Norcross (1971).

    a_lambda_3 : ``numpy.ndarray``
        Cross-section in function of wavelength and in unit of cm ** 2.
    """
    # The photoionization cross-section of He triplet is hard-coded with the
    # values calculated by Norcross 1971 (ADS:1971JPhB....4..652N). The
    # differential oscillator strength is calculated for bins of wavelength that
    # are not necessarily the same as the stellar spectrum wavelength bins.

    # Helium 2^3S differential oscillator strength
    data_array = np.array([[2593.01,     0.605],
                           [2528.27,     0.589],
                           [2275.74,     0.537],
                           [2023.15,     0.501],
                           [1655.63,     0.435],
                           [1214.41,     0.247],
                           [958.87,     0.1572],
                           [792.18,     0.1138],
                           [674.86,     0.0780],
                           [587.81,     0.0620],
                           [520.65,     0.0557],
                           [467.27,     0.0461],
                           [423.81,     0.0358],
                           [387.75,     0.0310],
                           [357.34,     0.0325],
                           [331.36,     0.0520],
                           [271.94,     0.343],
                           [271.21,     0.338],
                           [256.70,     0.274],
                           [243.01,     0.231],
                           [230.71,     0.200],
                           [219.59,     0.1750],
                           [209.49,     0.1537]
                           ])
    wavelength = np.flip(data_array[:, 0])
    differential_oscillator_strength = np.flip(data_array[:, 1])
    a_lambda_3 = 8.0670E-18 * differential_oscillator_strength
    return wavelength, a_lambda_3


# Collisional strengths for He
def he_collisional_strength():
    """
    Returns a hard-coded array containing the helium collisional strengths in
    function of temperature.

    Returns
    -------
    array : ``numpy.ndarray``
        Collisional strengths array. Columns: 0 = temperature, 1 = gamma_13,
        2 = gamma_31a, 3 = gamma_31b.
    """
    # log(T) [K]    gamma_13    gamma_31a   gamma_31b
    array = np.array([
        [3.75,            6.198E-2,    2.389,       7.965E-1],
        [4.00,            6.458E-2,    2.456,       9.579E-1],
        [4.25,            6.387E-2,    2.275,       1.042],
        [4.50,            6.157E-2,    1.916,       1.015],
        [4.75,            5.832E-2,    1.496,       8.950E-1],
        [5.00,            5.320E-2,    1.111,       7.265E-1],
        [5.25,            4.787E-2,    8.003E-1,    5.516E-1],
        [5.50,            4.018E-2,    5.660E-1,    3.948E-1],
        [5.75,            3.167E-2,    3.944E-1,    2.677E-1],
    ])
    return array


# Parametrized photoionization cross-section of atoms and ions from Verner+1996
def general_cross_section(wavelength, species):
    """
    Calculates the photoionization cross-section of an atomic or ion species
    using the parametrization of Verner et al. (1996)
    [https://ui.adsabs.harvard.edu/abs/1996ApJ...465..487V/abstract].

    Parameters
    ----------
    wavelength : ``float`` or ``numpy.ndarray``
        Photon wavelength in angstrom.

    species : ``str``
        String containing the species for which you request the cross-section in
        the format ``'X N'`` where ``X`` is the element and ``N`` is the
        ionization number. Example: ``'C II'``.

    Returns
    -------
    cross_section : ``float`` or ``numpy.ndarray``
        Cross-section in unit of cm ** (-2)
    """
    # Convert wavelength to energy
    energy = (c.h * c.c / wavelength / u.angstrom).to(u.eV).value

    parameters_dict = sigma_properties_v1996()
    energy_threshold, energy_max, energy_0, sigma_0, y_a, p, y_w, y_0, y_1 = \
        parameters_dict[species]

    # Check if the requested energy is within the energy range reported by
    # Verner et al. (1996)
    if isinstance(energy, float):
        if energy < energy_threshold or energy > energy_max:
            warn('The requested energy is outside the range of validity: '
                 '[{:.1f}, {:.1f}] eV'.format(energy_threshold, energy_max))
        else:
            pass
    elif isinstance(energy, np.ndarray):
        if energy[0] < energy_threshold or energy[-1] > energy_max:
            warn('The requested energy is outside the range of validity: '
                 '[{:.1f}, {:.1f}] eV'.format(energy_threshold, energy_max))
        else:
            pass
    else:
        raise ValueError('`energy` must be either `float` or `numpy.ndarray`.')

    mb = 1E-18  # cm ** (-2)
    x = energy / energy_0 - y_0
    y = (x ** 2 + y_1 ** 2) ** 0.5

    term1 = (x - 1) ** 2 + y_w ** 2
    term2 = y ** (0.5 * p - 5.5)
    term3 = (1 + (y / y_a) ** 0.5) ** (-p)
    function_y = term1 * term2 * term3

    cross_section = sigma_0 * function_y * mb  # cm ** (-2)
    cross_section[energy < energy_threshold] = 0.0

    return cross_section


# Some hard-coding to calculate the photoionization cross-sections from Verner
# et al. 1996.
def sigma_properties_v1996():
    """
    Function that hard-codes the cross-section parameters from Verner et al.
    (1996). We include only the species important for exoplanetary atmospheres.

    Returns
    -------
    parameters_dict : ``dict``
        Dictionary containing the parameters.
    """
    parameters_dict = {
        #         E_thresold, E_max, energy_0, sigma_0, y_a, p,
        #         y_w, y_0, y_1
        'C I':    [1.126E1, 2.910E2, 2.144E0, 5.027E2, 6.126E1, 5.101E0,
                   9.157E-2, 1.133E0, 1.607E0],
        'C II':   [2.438E1, 3.076E2, 4.058E-1, 8.709E0, 1.261E2, 8.578E0,
                   2.093E0, 4.929E1, 3.234E0],
        'C III':  [4.789E1, 3.289E2, 4.614E0, 1.539E4, 1.737E0, 1.593E1,
                   5.922E0, 4.378E-3, 2.528E-2],
        'N I':    [1.453E1, 4.048E2, 4.034E0, 8.235E2, 8.033E1, 3.928E0,
                   9.097E-2, 8.598E-1, 2.325E0],
        'O I':    [1.362E1, 5.380E2, 1.240E0, 1.745E3, 3.784E0, 1.764E1,
                   7.589E-2, 8.698E0, 1.271E-1],
        'O II':   [3.512E1, 5.581E2, 1.386E0, 5.967E1, 3.175E1, 8.943E0,
                   1.934E-2, 2.131E1, 1.503E-2],
        'O III':  [5.494E1, 5.840E2, 1.723E-1, 6.753E2, 3.852E2, 6.822E0,
                   1.191E-1, 3.839E-3, 4.569E-1],
        'Mg I':   [7.646E0, 5.490E1, 1.197E1, 1.372E8, 2.228E-1, 1.574E1,
                   2.805E-1, 0, 0],
        'Mg II':  [1.504E1, 6.569E1, 8.139E0, 3.278E0, 4.341E7, 3.610E0,
                   0, 0, 0],
        'Si I':   [8.152E0, 1.060E2, 2.317E1, 2.506E1, 2.057E1, 3.546E0,
                   2.837E-1, 1.672E-5, 4.207E-1],
        'Si II':  [1.635E1, 1.186E2, 2.556E0, 4.140E0, 1.337E1, 1.191E1,
                   1.570E0, 6.634E0, 1.272E-1],
        'Si III': [3.349E1, 1.311E2, 1.659E-1, 5.790E-4, 1.474E2, 1.336E1,
                   8.626E-1, 9.613E1, 6.442E-1],
        'Si IV':  [4.514E1, 1.466E2, 1.288E1, 6.083E0, 1.356E6, 3.353E0,
                   0, 0, 0],
        'Ca I':   [6.113E0, 3.443E1, 1.278E1, 5.370E5, 3.162E-1, 1.242E1,
                   4.477E-1, 1.012E-3, 1.851E-2],
        'Ca II':  [1.187E1, 4.090E1, 1.553E1, 1.064E7, 7.790E-1, 2.130E1,
                   6.453E-1, 2.161E-3, 6.706E-2],
        'Fe I':   [7.902E0, 6.600E1, 5.461E-2, 3.062E-1, 2.671E7, 7.923E0,
                   2.069E1, 1.382E2, 2.481E-1],
        'Fe II':  [1.619E1, 7.617E1, 1.761E-1, 4.365E3, 6.298E3, 5.204E0,
                   1.141E1, 9.272E1, 1.075E2],
    }
    return parameters_dict



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module computes the neutral and ionized populations of C in the upper
atmosphere.
"""

from __future__ import (division, print_function, absolute_import,
                        unicode_literals)
import numpy as np
import astropy.units as u
import astropy.constants as c
from scipy.integrate import simps, solve_ivp, odeint
from p_winds import tools, microphysics
import warnings


__all__ = ["radiative_processes", "electron_impact_ionization", "recombination",
           "charge_transfer", "ion_fraction"]


# Some hard coding based on the astrophysical literature
_SOLAR_CARBON_ABUNDANCE_ = 8.43  # Asplund et al. 2009
_SOLAR_CARBON_FRACTION_ = 10 ** (_SOLAR_CARBON_ABUNDANCE_ - 12.00)
_SOLAR_SILICON_ABUNDANCE_ = 7.51  # Asplund et al. 2009
_SOLAR_SILICON_FRACTION_ = 10 ** (_SOLAR_SILICON_ABUNDANCE_ - 12.00)


# Photoionization of C I (neutral) and C II (singly-ionized)
def radiative_processes(spectrum_at_planet):
    """
    Calculate the photoionization rate of carbon at null optical depth based
    on the EUV spectrum arriving at the planet.

    Parameters
    ----------
    spectrum_at_planet : ``dict``
        Spectrum of the host star arriving at the planet covering fluxes at
        least up to the wavelength corresponding to the energy to ionize
        carbon (11.26 eV, or 1101 Angstrom).

    Returns
    -------
    phi_ci : ``float``
        Ionization rate of C I at null optical depth in unit of 1 / s.

    phi_cii : ``float``
        Ionization rate of C II at null optical depth in unit of 1 / s.

    a_ci : ``float``
        Flux-averaged photoionization cross-section of C I in unit of cm ** 2.

    a_cii : ``float``
        Flux-averaged photoionization cross-section of C II in unit of cm ** 2.

    a_h_ci : ``float``
        Flux-averaged photoionization cross-section of H I in the range absorbed
        by C I in unit of cm ** 2.

    a_h_cii : ``float``
        Flux-averaged photoionization cross-section of H I in the range absorbed
        by C II in unit of cm ** 2.

    a_he : ``float``
        Flux-averaged photoionization cross-section of He I in unit of cm ** 2.
    """
    wavelength = (spectrum_at_planet['wavelength'] *
                  spectrum_at_planet['wavelength_unit']).to(u.angstrom).value
    flux_lambda = (spectrum_at_planet['flux_lambda'] * spectrum_at_planet[
        'flux_unit']).to(u.erg / u.s / u.cm ** 2 / u.angstrom).value
    energy = ((c.h * (c.c / wavelength / u.angstrom).to(u.Hz)).to(u.eV)).value
    energy_erg = (energy * u.eV).to(u.erg).value

    # Auxiliary definitions
    parameters_dict = microphysics.sigma_properties_v1996()
    energy_threshold_ci = parameters_dict['C I'][0]  # Ionization threshold in
    # eV
    energy_threshold_cii = parameters_dict['C II'][0]  # Ionization threshold in
    # eV
    wl_break_ci = 12398.42 / energy_threshold_ci  # C I ionization threshold in
    # angstrom
    wl_break_cii = 12398.42 / energy_threshold_cii  # C II ionization threshold
    # in angstrom
    wl_break_he = 504  # He ionization threshold in angstrom
    i0 = tools.nearest_index(wavelength, wl_break_he)
    i1 = tools.nearest_index(wavelength, wl_break_ci)
    i2 = tools.nearest_index(wavelength, wl_break_cii)
    wavelength_cut_0 = wavelength[:i0]
    flux_lambda_cut_0 = flux_lambda[:i0]
    wavelength_cut_1 = wavelength[:i1]
    flux_lambda_cut_1 = flux_lambda[:i1]
    energy_cut_1 = energy_erg[:i1]
    wavelength_cut_2 = wavelength[:i2]
    flux_lambda_cut_2 = flux_lambda[:i2]
    energy_cut_2 = energy_erg[:i2]

    # Calculate the photoionization cross-section
    a_lambda_ci = microphysics.general_cross_section(wavelength_cut_1,
                                                     species='C I')
    a_lambda_cii = microphysics.general_cross_section(wavelength_cut_2,
                                                      species='C II')

    # The flux-averaged photoionization cross-section of C I and C II
    a_ci = abs(simps(flux_lambda_cut_1 * a_lambda_ci, wavelength_cut_1) /
               simps(flux_lambda_cut_1, wavelength_cut_1))
    a_cii = abs(simps(flux_lambda_cut_2 * a_lambda_cii, wavelength_cut_2) /
                simps(flux_lambda_cut_2, wavelength_cut_2))

    # The flux-averaged photoionization cross-section of H is also going to be
    # needed because it adds to the optical depth that C I and C II see.
    a_lambda_h_ci = microphysics.hydrogen_cross_section(
        wavelength=wavelength_cut_1)
    a_h_ci = abs(simps(flux_lambda_cut_1 * a_lambda_h_ci, wavelength_cut_1) /
                 simps(flux_lambda_cut_1, wavelength_cut_1))
    a_lambda_h_cii = microphysics.hydrogen_cross_section(
        wavelength=wavelength_cut_2)
    a_h_cii = abs(simps(flux_lambda_cut_2 * a_lambda_h_cii, wavelength_cut_2) /
                  simps(flux_lambda_cut_2, wavelength_cut_2))

    # Same for the He atoms, but only up to the He ionization threshold
    a_lambda_he = microphysics.helium_total_cross_section(wavelength_cut_0)
    a_he = abs(simps(flux_lambda_cut_0 * a_lambda_he, wavelength_cut_0) /
               simps(flux_lambda_cut_0, wavelength_cut_0))

    # Calculate the photoionization rates
    phi_ci = abs(simps(flux_lambda_cut_1 * a_lambda_ci / energy_cut_1,
                 wavelength_cut_1))
    phi_cii = abs(simps(flux_lambda_cut_2 * a_lambda_cii / energy_cut_2,
                        wavelength_cut_2))

    return phi_ci, phi_cii, a_ci, a_cii, a_h_ci, a_h_cii, a_he


# Ionization rate of C by electron impact
def electron_impact_ionization(electron_temperature):
    """
    Calculates the electron impact ionization rate that consumes neutral C and
    produces singly-ionized C. Based on the formula of Voronov 1997
    (https://ui.adsabs.harvard.edu/abs/1997ADNDT..65....1V/abstract).

    Parameters
    ----------
    electron_temperature : ``float``
        Temperature of the plasma where the electrons are embedded in unit of
        Kelvin.

    Returns
    -------
    ionization_rate_ci : ``float``
        Ionization rate of neutral C into singly-ionized C in unit of
        cm ** 3 / s.

    ionization_rate_cii : ``float``
        Ionization rate of singly-ionized C into doubly-ionized C in unit of
        cm ** 3 / s.
    """
    boltzmann_constant = 8.617333262145179e-05  # eV / K
    electron_energy = boltzmann_constant * electron_temperature
    energy_ratio_ci = 11.3 / electron_energy
    energy_ratio_cii = 24.4 / electron_energy
    ionization_rate_ci = 6.85E-8 * (0.193 + energy_ratio_ci) ** (-1) * \
        energy_ratio_ci ** 0.25 * np.exp(-energy_ratio_ci)
    ionization_rate_cii = 1.86E-8 * (0.286 + energy_ratio_cii) ** (-1) * \
        energy_ratio_cii ** 0.24 * np.exp(-energy_ratio_cii)
    return ionization_rate_ci, ionization_rate_cii


# Recombination of singly-ionized C into neutral C
def recombination(electron_temperature):
    """
    Calculates the rate of recombination of singly-ionized C with an electron to
    produce a neutral C atom. Based on the formulation of Woodall et al. 2007
    (https://ui.adsabs.harvard.edu/abs/2007A%26A...466.1197W/abstract). Also
    calculates the recombination of doubly-ionized C with an electron to produce
    a singly-ionized C ion. Based on the formulation of Aldrovandi & Péquignot
    1973 (https://ui.adsabs.harvard.edu/abs/1973A%26A....25..137A/abstract).

    Parameters
    ----------
    electron_temperature : ``float``
        Temperature of the plasma where the electrons are embedded in unit of
        Kelvin.

    Returns
    -------
    alpha_rec_ci : ``float``
        Recombination rate of C II into C I in units of cm ** 3 / s.

    alpha_rec_cii : ``float``
        Recombination rate of C III into C II in units of cm ** 3 / s.
    """
    alpha_rec_ci = 4.67E-12 * (300 / electron_temperature) ** 0.60
    alpha_rec_cii = 2.3E-12 * (1000 / electron_temperature) ** 0.645
    return alpha_rec_ci, alpha_rec_cii


# Charge transfer between C and H, He and Si
def charge_transfer(temperature):
    """
    Calculates the charge exchange rates of C with H, He and Si nuclei. Based on
    the formulation of Stancil et al. 1998
    (https://ui.adsabs.harvard.edu/abs/1998ApJ...502.1006S/abstract),
    Woodall et al. 2007
    (https://ui.adsabs.harvard.edu/abs/2007A%26A...466.1197W/abstract),
    Glover & Jappsen 2007
    (https://ui.adsabs.harvard.edu/abs/2007ApJ...666....1G/abstract),
    Kingdon & Ferland 1996
    (https://ui.adsabs.harvard.edu/abs/1996ApJS..106..205K/abstract), and
    Brown 1972 (https://ui.adsabs.harvard.edu/abs/1972ApJ...174..511B/abstract).

    Parameters
    ----------
    temperature : ``float``
        Isothermal temperature of the upper atmosphere in unit of Kelvin.

    Returns
    -------
    ct_rate_ci_hp : ``float``
        Charge transfer rate between neutral C and H+ in units of cm ** 3 / s.

    ct_rate_cii_h : ``float``
        Charge transfer rate between C+ and neutral H in units of cm ** 3 / s.

    ct_rate_ci_hep : ``float``
        Charge transfer rate between neutral C and He+ in units of cm ** 3 / s.

    ct_rate_cii_sii : ``float``
        Charge transfer rate between C+ and neutral Si in units of cm ** 3 / s.

    ct_rate_ciii_h : ``float``)
        Charge transfer rate between C++ and neutral H in units of cm ** 3 / s.

    ct_rate_ciii_he : ``float``)
        Charge transfer rate between C++ and neutral He in units of cm ** 3 / s.
    """
    # Recombination of C II into C I
    ct_rate_cii_h = 6.30E-17 * (300 / temperature) ** (-1.96) * \
        np.exp(-1.7E5 / temperature)
    ct_rate_cii_sii = 2.1E-9

    # Ionization of C I into C II
    ct_rate_ci_hp = 1.31E-15 * (300 / temperature) ** (-0.213)
    ct_rate_ci_hep = 2.5E-15 * (300 / temperature) ** (-1.597)

    # Recombination of C III into C II
    ct_rate_ciii_h = 1.67E-4 * (temperature / 10000) ** 2.79 * \
        (1 + 304.72 * np.exp(-4.07 * temperature / 10000))
    ct_rate_ciii_he = 1.23E-9  # Very approximated from Brown 1972, but should
    # be good enough for temperatures near 10,000 K

    return ct_rate_ci_hp, ct_rate_cii_h, ct_rate_ci_hep, ct_rate_cii_sii, \
        ct_rate_ciii_h, ct_rate_ciii_he


# Calculation the number fractions of C II and C III
def ion_fraction(radius_profile, velocity, density, hydrogen_ion_fraction,
                 helium_ion_fraction, planet_radius, temperature, h_fraction,
                 speed_sonic_point, radius_sonic_point, density_sonic_point,
                 spectrum_at_planet, c_fraction=_SOLAR_CARBON_FRACTION_,
                 initial_f_c_ion=np.array([0.0, 0.0]), relax_solution=False,
                 convergence=0.01, max_n_relax=10, method='odeint',
                 return_rates=False, **options_solve_ivp):
    """
    Calculates the fractions of singly- and doubly-ionized carbon in the upper
    atmosphere in function of the radius in unit of planetary radius.

    Parameters
    ----------
    radius_profile : ``numpy.ndarray``
        Radius in unit of planetary radii.

    velocity : ``numpy.ndarray``
         Velocities sampled at the values of ``radius_profile`` in units of
         sound speed. Similar to the output of ``parker.structure()``.

    density : ``numpy.ndarray``
        Densities sampled at the values of ``radius_profile`` in units of
        density at the sonic point. Similar to the output of
        ``parker.structure()``.

    hydrogen_ion_fraction : ``numpy.ndarray``
        Number fraction of H ion over total H in the upper atmosphere in
        function of radius. Similar to the output of
        ``hydrogen.ion_fraction()``.

    helium_ion_fraction : ``numpy.ndarray``
        Number fraction of He ion over total He in the upper atmosphere in
        function of radius. Similar to the output of
        ``helium.population_fraction()``, but should be ``1 - f_1_r - f_3_r``.

    planet_radius : ``float``
        Planetary radius in unit of Jupiter radius.

    temperature : ``float``
        Isothermal temperature of the upper atmosphere in unit of Kelvin.

    h_fraction : ``float``
        Total (ion + neutral) H number fraction of the atmosphere.

    speed_sonic_point : ``float``
        Speed of sound in the outflow in units of km / s.

    radius_sonic_point : ``float``
        Radius of the sonic point in unit of Jupiter radius.

    density_sonic_point : ``float``
        Density at the sonic point in units of g / cm ** 3.

    spectrum_at_planet : ``dict``
        Spectrum of the host star arriving at the planet covering fluxes at
        least up to the wavelength corresponding to the energy to ionize
        carbon (11.26 eV, or 1101 Angstrom). Can be generated using
        ``tools.make_spectrum_dict``.

    c_fraction : ``float``, optional
        Fraction of total carbon in the upper atmosphere. Default value assumes
        solar abundance.

    initial_f_c_ion : ``numpy.ndarray``, optional
        The initial ion fractions are the `y0` of the differential equation to
        be solved. This array has two items: the initial fraction of
        singly-ionized and doubly-ionized carbon in the inner layer of the
        atmosphere. The default value for this parameter is
        ``numpy.array([0.0, 0.0])``, i.e., fully neutral at the inner layer.

    relax_solution : ``bool``, optional
        The first solution is calculating by initially assuming the entire
        atmosphere is in neutral state. If ``True``, the solution will be
        re-calculated in a loop until it converges to a delta_f of 1%, or for a
        maximum of 10 loops (default parameters). Default is ``False``.

    convergence : ``float``, optional
        Value of delta_f at which to stop the relaxation of the solution for
        ``f_r``. Default is 0.01.

    max_n_relax : ``int``, optional
        Maximum number of loops to perform the relaxation of the solution for
        the ion fractions. Default is 10.

    method : ``str``, optional
        If method is ``'odeint'``, then ``scipy.integrate.odeint()`` is used
        instead of ``scipy.integrate.solve_ivp()`` to calculate the steady-state
        distribution of helium. The first seems to be at least twice faster than
        the second in some situations. Any other method will fall back to an
        option of ``solve_ivp()`` methods. For example, if ``method`` is set to
        ``'Radau'``, then use ``solve_ivp(method='Radau')``. Default is
        ``'odeint'``.

    return_rates : ``bool``, optional
        If ``True``, then this function also returns a ``dict`` object
        containing the various reaction rates in function of radius and in units
        of 1 / s. Default is ``False``.

    **options_solve_ivp:
        Options to be passed to the ``scipy.integrate.solve_ivp()`` solver. You
        may want to change the options ``atol`` (absolute tolerance; default is
        1E-6) or ``rtol`` (relative tolerance; default is 1E-3). If you are
        having numerical issues, you may want to decrease the tolerance by a
        factor of 10 or 100, or 1000 in extreme cases.

    Returns
    -------
    f_cii_r : ``numpy.ndarray``
        Fraction of singly-ionized carbon in function of radius.

    f_ciii_r : ``numpy.ndarray``
        Fraction of doubly-ionized carbon in function of radius.

    reaction_rates : ``dict``
        Dictionary containing the reaction rates in function of radius and in
        units of 1 / s. Only returned when ``return_rates`` is set to ``True``.
        Here is a short description of the dict keys:

        * `ionization_CI`: Photoionization of C I into C II
        * `ionization_CII`: Photoionization of C II into C III
        * `recombination_CII`: Recombination of C II into C I
        * `recombination_CIII`: Recombination of C III into C II
        * `e_impact_ion_CI`: Electron impact ionization of C I into C II
        * `e_impact_ion_CII`: Electron impact ionization of C II into C III
        * `charge_exchange_CI_HII`: Charge exchange between C I and H II
        * `charge_exchange_CI_HeII`: Charge exchange between C I and He II
        * `charge_exchange_CII_HI`: Charge exchange between C II and H I
        * `charge_exchange_CIII_HI`: Charge exchange between C III and H I
        * `charge_exchange_CIII_HeI`: Charge exchange between C III and He I
    """
    vs = speed_sonic_point  # km / s
    rs = radius_sonic_point  # jupiterRad
    rhos = density_sonic_point  # g / cm ** 3

    # Recombination rates of C in unit of rs ** 2 * vs
    alpha_rec_unit = ((rs * 7.1492E+09) ** 2 * vs * 1E5)  # cm ** 3 / s
    alpha_rec_ci, alpha_rec_cii = recombination(temperature)
    alpha_rec_ci = alpha_rec_ci / alpha_rec_unit
    alpha_rec_cii = alpha_rec_cii / alpha_rec_unit

    # Hydrogen mass in unit of rhos * rs ** 3
    m_h_unit = (rhos * (rs * 7.1492E+09) ** 3)  # Converted to g
    m_h = 1.67262192E-24 / m_h_unit

    # Photoionization rates at null optical depth at the distance of the planet
    # from the host star, in unit of vs / rs, and the flux-averaged
    # cross-sections in units of rs ** 2
    phi_unit = vs * 1E5 / rs / 7.1492E+09  # 1 / s
    phi_ci, phi_cii, a_ci, a_cii, a_h_ci, a_h_cii, a_he = radiative_processes(
        spectrum_at_planet)
    phi_ci = phi_ci / phi_unit
    a_ci = a_ci / (rs * 7.1492E+09) ** 2
    a_h_ci = a_h_ci / (rs * 7.1492E+09) ** 2
    phi_cii = phi_cii / phi_unit
    a_cii = a_cii / (rs * 7.1492E+09) ** 2
    a_h_cii = a_h_cii / (rs * 7.1492E+09) ** 2
    a_he = a_he / (rs * 7.1492E+09) ** 2

    # Electron-impact ionization rate for C I in the same unit as the
    # recombination rates
    ionization_rate_ci, ionization_rate_cii = \
        electron_impact_ionization(temperature)
    ionization_rate_ci = ionization_rate_ci / alpha_rec_unit
    ionization_rate_cii = ionization_rate_cii / alpha_rec_unit

    # Charge transfer rates in the same unit as the recombination rates
    ct_rate_ci_hp, ct_rate_cii_h, ct_rate_ci_hep, ct_rate_cii_sii, \
        ct_rate_ciii_h, ct_rate_ciii_he = charge_transfer(temperature)
    ct_rate_cii_h = ct_rate_cii_h / alpha_rec_unit
    ct_rate_ci_hp = ct_rate_ci_hp / alpha_rec_unit
    ct_rate_ci_hep = ct_rate_ci_hep / alpha_rec_unit
    ct_rate_ciii_h = ct_rate_ciii_h / alpha_rec_unit
    ct_rate_ciii_he = ct_rate_ciii_he / alpha_rec_unit

    # We solve the steady-state ionization balance in a similar way that we do
    # for He

    # The radius in unit of radius at the sonic point
    r = radius_profile * planet_radius / rs
    dr = np.diff(r)
    dr = np.concatenate((dr, np.array([dr[-1], ])))

    # With all this setup done, now we need to assume something about the
    # distribution of neutral C in the atmosphere. Let's assume it based on the
    # initial guess input.
    column_density = np.flip(np.cumsum(np.flip(dr * density)))  # Total column
                                                                # density
    column_density_h_0 = np.flip(  # Column density of atomic H only
        np.cumsum(np.flip(dr * density * (1 - hydrogen_ion_fraction))))
    he_fraction = 1 - h_fraction
    column_density_he_0 = np.flip(  # Column density of atomic He only
        np.cumsum(np.flip(dr * density * he_fraction *
                          (1 - helium_ion_fraction))))
    k1 = h_fraction / (h_fraction + 4 * he_fraction + 12 * c_fraction) / m_h
    k2 = he_fraction / (h_fraction + 4 * he_fraction + 12 * c_fraction) / m_h
    k3 = c_fraction / (h_fraction + 4 * he_fraction + 12 * c_fraction) / m_h
    tau_ci_h = k1 * a_h_ci * column_density_h_0
    tau_cii_h = k1 * a_h_cii * column_density_h_0
    tau_c_he = k2 * a_he * column_density_he_0
    tau_ci = \
        (1 - initial_f_c_ion[0] - initial_f_c_ion[1]) * k3 * a_ci * \
        column_density + tau_ci_h + tau_c_he
    tau_cii = \
        initial_f_c_ion[0] * k3 * a_cii * column_density + tau_cii_h + tau_c_he

    # The differential equation
    def _fun(_r, y, rates=False):
        f_cii = y[0]
        f_ciii = y[1]

        _v = np.interp(_r, r, velocity)
        _rho = np.interp(_r, r, density)
        f_h_ion = np.interp(_r, r, hydrogen_ion_fraction)  # Fraction of H+
        f_he_ion = np.interp(_r, r, helium_ion_fraction)  # Fraction of He+

        # Assume the number density of electrons is equal to the number density
        # of H ions + He ions
        # Since we may run into loss of numerical precision here (big numbers),
        # we manipulate the equations to avoid this problem. It looks a bit
        # messy, but it is necessary
        log_term_1 = np.log(k1) + np.log(_rho)  # H ions
        log_term_2 = np.log(k2) + np.log(_rho)  # He ions
        ionization_rate_ci_n_e = \
            np.exp(log_term_1 + np.log(ionization_rate_ci)) * f_h_ion + \
            np.exp(log_term_2 + np.log(ionization_rate_ci)) * f_he_ion
        ct_rate_ci_hp_n_h_plus = \
            np.exp(log_term_1 + np.log(ct_rate_ci_hp)) * f_h_ion
        ct_rate_ci_hep_n_he_plus = \
            np.exp(log_term_2 + np.log(ct_rate_ci_hep)) * f_he_ion
        alpha_rec_ci_n_e = \
            np.exp(log_term_1 + np.log(alpha_rec_ci)) * f_h_ion + \
            np.exp(log_term_2 + np.log(alpha_rec_ci)) * f_he_ion
        ct_rate_cii_h_n_h0 = \
            np.exp(log_term_1 + np.log(ct_rate_cii_h)) * (1 - f_h_ion)
        alpha_rec_cii_n_e = \
            np.exp(log_term_1 + np.log(alpha_rec_cii)) * f_h_ion + \
            np.exp(log_term_2 + np.log(alpha_rec_cii)) * f_he_ion
        ct_rate_ciii_h_n_h0 = \
            np.exp(log_term_1 + np.log(ct_rate_ciii_h)) * (1 - f_h_ion)
        ct_rate_ciii_he_n_he0 = \
            np.exp(log_term_2 + np.log(ct_rate_ciii_he)) * (1 - f_he_ion)
        ionization_rate_cii_n_e = \
            np.exp(log_term_1 + np.log(ionization_rate_cii)) * f_h_ion + \
            np.exp(log_term_2 + np.log(ionization_rate_cii)) * f_he_ion

        # Terms of dfcii_dr
        t_ci = np.interp(_r, r, tau_ci)
        term11 = (1 - f_cii - f_ciii) * phi_ci * np.exp(-t_ci)  # Photo-
        # ionization
        term12 = (1 - f_cii - f_ciii) * ionization_rate_ci_n_e  # Electron-
        # impact ionization
        term13 = (1 - f_cii - f_ciii) * ct_rate_ci_hp_n_h_plus  # Charge
        # exchange with H+
        term14 = (1 - f_cii - f_ciii) * ct_rate_ci_hep_n_he_plus  # Charge
        # exchange with He+
        term15 = f_cii * alpha_rec_ci_n_e  # Recombination of C II into C I
        term16 = f_cii * ct_rate_cii_h_n_h0  # Charge exchange of C II with
        # neutral H
        term17 = f_ciii * alpha_rec_cii_n_e  # Recombination of C III into
        # C II
        term18 = f_ciii * ct_rate_ciii_h_n_h0  # Charge exchange of C III with
        # neutral H
        term19 = f_ciii * ct_rate_ciii_he_n_he0  # Charge exchange of C III
        # with neutral He

        # Terms of dfciii_dr
        t_cii = np.interp(_r, r, tau_cii)
        term21 = f_cii * phi_cii * np.exp(-t_cii)  # Photoionization
        term22 = f_cii * ionization_rate_cii_n_e  # Electron-impact ionization

        dfcii_dr = (term11 + term12 + term13 + term14 - term15 - term16 +
                    term17 + term18 + term19 - term21 - term22) / _v
        dfciii_dr = (term21 + term22 - term17 - term18 - term19) / _v

        if rates is False:
            return np.array([dfcii_dr, dfciii_dr])
        else:
            return np.array([term11, term21, term15, term17, term12, term22,
                             term13, term14, term16, term18, term19]) * phi_unit

    if method == 'odeint':
        # Since 'odeint' yields only warnings when precision is lost or when
        # there is a problem, we transform these warnings into an exception
        with warnings.catch_warnings():
            warnings.filterwarnings("error")
            try:
                sol = odeint(_fun, y0=initial_f_c_ion, t=r, tfirst=True, )
            except Warning:
                raise RuntimeError('The solver ``odeint`` failed to obtain a '
                                   'solution.')
        f_cii_r = np.copy(sol).T[0]
        f_ciii_r = np.copy(sol).T[1]
    else:
        # We solve it using `scipy.solve_ivp`
        sol = solve_ivp(_fun, (r[0], r[-1],), initial_f_c_ion, t_eval=r,
                        method=method, **options_solve_ivp)
        f_cii_r = sol['y'][0]
        f_ciii_r = sol['y'][1]
        # When `solve_ivp` has problems, it may return an array with different
        # size than `r`. So we raise an exception if this happens
        if len(f_cii_r) != len(r) or len(f_ciii_r) != len(r):
            raise RuntimeError('The solver ``solve_ivp`` failed to obtain a'
                               ' solution.')

    # High densities can be numerically unstable and produce unphysical values
    # of `f_r`, so we replace negative values with zero and values above 1.0
    # with 1.0
    f_cii_r[f_cii_r < 0] = 1E-15
    f_ciii_r[f_ciii_r < 0] = 1E-15
    f_cii_r[f_cii_r > 1.0] = 1.0
    f_ciii_r[f_ciii_r > 1.0] = 1.0

    # For the sake of self-consistency, there is the option of repeating the
    # calculation of f_r by updating the optical depth with the new ion
    # fractions.
    if relax_solution is True:
        for i in range(max_n_relax):
            previous_f_cii_r = np.copy(f_cii_r)
            previous_f_ciii_r = np.copy(f_ciii_r)

            # Re-calculate the column densities
            tau_ci = \
                k3 * a_ci * np.flip(np.cumsum(
                    np.flip(dr * density * (1 - f_cii_r - f_ciii_r)))) + \
                tau_ci_h + tau_c_he
            tau_cii = k3 * a_cii * np.flip(
                np.cumsum(np.flip(dr * density * f_cii_r))) + tau_cii_h + \
                tau_c_he

            # Solve it again
            if method == 'odeint':
                sol = odeint(_fun, y0=initial_f_c_ion, t=r, tfirst=True)
                f_cii_r = np.copy(sol).T[0]
                f_ciii_r = np.copy(sol).T[1]
            else:
                sol = solve_ivp(_fun, (r[0], r[-1],), initial_f_c_ion,
                                t_eval=r,
                                method=method, **options_solve_ivp)
                f_cii_r = sol['y'][0]
                f_ciii_r = sol['y'][1]

            # Replace negative values with zero and values above 1.0 with
            # 1.0
            f_cii_r[f_cii_r < 0] = 1E-15
            f_ciii_r[f_ciii_r < 0] = 1E-15
            f_cii_r[f_cii_r > 1.0] = 1.0
            f_ciii_r[f_ciii_r > 1.0] = 1.0

            # Calculate the relative change of f_ion in the outer shell of
            # the atmosphere (where we expect the most important change)
            relative_delta_f_cii = abs(np.sum(f_cii_r - previous_f_cii_r)) \
                / np.sum(previous_f_cii_r)
            relative_delta_f_ciii = \
                abs(np.sum(f_ciii_r - previous_f_ciii_r)) / \
                np.sum(previous_f_ciii_r)

            # Break the loop if convergence is achieved
            if relative_delta_f_cii < convergence and \
                    relative_delta_f_ciii < convergence:
                break
            else:
                pass
    else:
        pass

    if return_rates is False:
        return f_cii_r, f_ciii_r
    else:
        ion_rate_ci, ion_rate_cii, recomb_rate_cii, recomb_rate_ciii, \
            e_imp_ion_ci, e_imp_ion_cii, ch_exchange_ci_hii, \
            ch_exchange_ci_heii, ch_exchange_cii_hi, ch_exchange_ciii_hi, \
            ch_exchange_ciii_hei = _fun(r, [f_cii_r, f_ciii_r], rates=True)
        reaction_rates = {'ionization_CI': ion_rate_ci,
                          'ionization_CII': e_imp_ion_cii,
                          'recombination_CII': recomb_rate_cii,
                          'recombination_CIII': recomb_rate_ciii,
                          'e_impact_ion_CI': e_imp_ion_ci,
                          'e_impact_ion_CII': e_imp_ion_cii,
                          'charge_exchange_CI_HII': ch_exchange_ci_hii,
                          'charge_exchange_CI_HeII': ch_exchange_ci_heii,
                          'charge_exchange_CII_HI': ch_exchange_cii_hi,
                          'charge_exchange_CIII_HI': ch_exchange_ciii_hi,
                          'charge_exchange_CIII_HeI': ch_exchange_ciii_hei}
        return f_cii_r, f_ciii_r, reaction_rates



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

if sys.argv[-1] == "publish":
    os.system("python setup.py sdist upload")
    sys.exit()

setup(
    name="p_winds",
    version="1.4.7",
    author="Leonardo dos Santos",
    author_email="ldsantos@stsci.edu",
    packages=["p_winds"],
    url="https://github.com/ladsantos/p-winds",
    license="MIT",
    description="Parker wind models for planetary atmospheres",
    install_requires=[line.strip() for line in
                      open('requirements.txt', 'r').readlines()],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ]
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-


from p_winds import parker


# The only slightly complicated code in the ``parker`` module is ``structure()``
# so let's test how it is working
# The velocity and density calculated by ``structure()`` should yield values of
# 1.0 at r = 1.0.
def test_structure(r=1.0, precision_threshold=1E-6):
    velocity, density = parker.structure(r)
    assert abs((velocity - 1.0) / velocity) < precision_threshold
    assert abs((density - 1.0) / density) < precision_threshold


# The sound speed for a gas made of 100% atomic hydrogen at 10,000 K should be
# 9.08537273 km /s
def test_sound_speed(temperature=10000, h_he=1.0,
                     precision_threshold=1E-6):
    vs = parker.sound_speed(temperature, h_he)
    assert abs((vs - 9.08537273) / vs) < precision_threshold
